"""Tests for the Tools module."""


def test_pcs_parser() -> None:
    """Tests for file pcs_parser."""
    # TODO: Write test
    pass


def test_general() -> None:
    """Tests for general file."""
    # TODO: Write test
    pass


def test_runsolver_parsing() -> None:
    """Tests for file runsolver_parsing."""
    # TODO: Write test
    pass


def test_solver_wrapper_parsing() -> None:
    """Tests for file solver_wrapper_parsing."""
    # TODO: Write test
    pass



"""Test files for Extractor class."""


def test_extractor_constructor() -> None:
    """Test for constructor."""
    # TODO: Write test
    pass


def test_features() -> None:
    """Test for property features."""
    # TODO: Write test
    pass


def test_feature_groups() -> None:
    """Test for property feature_groups."""
    # TODO: Write test
    pass


def test_output_dimension() -> None:
    """Test for property output_dimension."""
    # TODO: Write test
    pass


def test_groupwise_computation() -> None:
    """Test for property groupwise_computation."""
    # TODO: Write test
    pass


def test_build_cmd() -> None:
    """Test for method build_cmd."""
    # TODO: Write test
    pass


def test_run() -> None:
    """Test for method run."""
    # TODO: Write test
    pass


def test_get_feature_vector() -> None:
    """Test for method get_feature_vector."""
    # TODO: Write test
    pass



import logging

import numpy as np
import pandas as pd

from sklearn.impute import SimpleImputer

from ConfigSpace.hyperparameters import CategoricalHyperparameter, \
    UniformFloatHyperparameter, UniformIntegerHyperparameter
from ConfigSpace import Configuration
from ConfigSpace.configuration_space import ConfigurationSpace

from autofolio.aslib_scenario import ASlibScenario

__author__ = "Marius Lindauer"
__license__ = "BSD"


class ImputerWrapper(object):

    @staticmethod
    def add_params(cs: ConfigurationSpace):
        '''
            adds parameters to ConfigurationSpace 
        '''
        stratgey = CategoricalHyperparameter(
            "imputer_strategy", choices=["mean", "median", "most_frequent"], default_value="mean")
        cs.add_hyperparameter(stratgey)

    def __init__(self):
        '''
            Constructor
        '''
        self.imputer = None
        self.active = False

        self.logger = logging.getLogger("MissingValueImputation")

    def fit(self, scenario: ASlibScenario, config: Configuration):
        '''
            fit pca object to ASlib scenario data

            Arguments
            ---------
            scenario: data.aslib_scenario.ASlibScenario
                ASlib Scenario with all data in pandas
            config: ConfigSpace.Configuration
                configuration
        '''

        self.imputer = SimpleImputer(strategy=config.get("imputer_strategy"))
        self.imputer.fit(scenario.feature_data.values)
        self.active = True

    def transform(self, scenario: ASlibScenario):
        '''
            transform ASLib scenario data

            Arguments
            ---------
            scenario: data.aslib_scenario.ASlibScenario
                ASlib Scenario with all data in pandas

            Returns
            -------
            data.aslib_scenario.ASlibScenario
        '''
        self.logger.debug("Impute Missing Feature Values")

        values = self.imputer.transform(
            np.array(scenario.feature_data.values))
        scenario.feature_data = pd.DataFrame(
            data=values, index=scenario.feature_data.index, columns=scenario.feature_data.columns)

        return scenario

    def fit_transform(self, scenario: ASlibScenario, config: Configuration):
        '''
            fit and transform

            Arguments
            ---------
            scenario: data.aslib_scenario.ASlibScenario
                ASlib Scenario with all data in pandas
            config: ConfigSpace.Configuration
                configuration

            Returns
            -------
            data.aslib_scenario.ASlibScenario
        '''
        self.fit(scenario, config)
        scenario = self.transform(scenario)
        return scenario
    
    def get_attributes(self):
        '''
            returns a list of tuples of (attribute,value) 
            for all learned attributes
            
            Arguments
            ---------
            config: ConfigSpace.Configuration
                configuration

            
            Returns
            -------
            list of tuples of (attribute,value) 
        '''
        return ["Strategy=%s" %(self.imputer.strategy)]


"""Tests for Validator Class."""


def test_validator_constructor() -> None:
    """Test Validator initialization."""
    # TODO: Write test
    pass


def test_validate() -> None:
    """Test Validator validate method."""
    # NOTE: Make sure runrunner calls are mocked to prevent slurm job submission
    # TODO: Write test
    pass


def test_retrieve_raw_results() -> None:
    """Test Validator retrieve_raw_results method."""
    # TODO: Write test
    pass


def test_get_validation_results() -> None:
    """Test Validator get_validation_results method."""
    # TODO: Write test
    pass


def test_append_entry_to_csv() -> None:
    """Test Validator append_entry_to_csv method."""
    # TODO: Write test
    pass



#!/usr/bin/env python3
"""Sparkle class to organise configuration output."""

from __future__ import annotations

from sparkle.platform import \
    generate_report_for_configuration as sgrfch
from sparkle.solver import Solver
from sparkle.instance import InstanceSet
from sparkle.configurator.configurator import Configurator, ConfigurationScenario
from sparkle.solver.validator import Validator
from sparkle.platform.output.structures import ValidationResults, ConfigurationResults
from sparkle.types import SolverStatus

import json
from pathlib import Path


class ConfigurationOutput:
    """Class that collects configuration data and outputs it a JSON format."""

    def __init__(self: ConfigurationOutput, path: Path, solver: Solver,
                 configurator: Configurator, instance_set_train: InstanceSet,
                 instance_set_test: InstanceSet, output: Path) -> None:
        """Initialize Configurator Output class.

        Args:
            path: Path to configuration output directory
            solver: Solver object
            configurator: The configurator that was used
            instance_set_train: Instance set used for training
            instance_set_test: Instance set used for testing
            output: Path to the output directory
        """
        self.solver = solver
        self.configurator = configurator
        self.instance_set_train = instance_set_train
        self.instance_set_test = instance_set_test
        self.directory = path
        self.output = output / "configuration.json" if not output.is_file() else output

        solver_dir_name = path.name
        scenario_file = path / f"{solver_dir_name}_scenario.txt"
        if not scenario_file.is_file():
            raise Exception("Can't find scenario file")

        # Sets scenario on configurator object
        self.configurator.scenario = \
            configurator.scenario_class.from_file(scenario_file, self.solver,
                                                  self.instance_set_train)
        self.configurator.scenario._set_paths(self.configurator.output_path)

        # Retrieve all configurations
        config_path = path / "validation" / "configurations.csv"
        self.configurations = self.get_configurations(config_path)

        # Retrieve best found configuration
        objective = self.configurator.scenario.sparkle_objective
        _, self.best_config = self.configurator.get_optimal_configuration(
            self.solver, self.instance_set_train, objective)

        # Retrieves validation results for all configurations
        self.validation_results = []
        for config in self.configurations:
            val_res = self.get_validation_data(self.instance_set_train,
                                               config)
            self.validation_results.append(val_res)

        # Retrieve test validation results if they exist
        if self.instance_set_test is not None:
            self.validation_results_test = []
            for config in self.configurations:
                val_res = self.get_validation_data(self.instance_set_test,
                                                   config)
                self.validation_results_test.append(val_res)

    def get_configurations(self: ConfigurationOutput, config_path: Path) -> list[dict]:
        """Read all configurations and transform them to dictionaries."""
        configs = []
        # Check if the path exists and is a file
        if config_path.exists() and config_path.is_file():
            with config_path.open("r") as file:
                for line in file:
                    config = Solver.config_str_to_dict(line.strip())
                    if config not in configs:
                        configs.append(config)
        return configs

    def get_validation_data(self: ConfigurationOutput, instance_set: InstanceSet,
                            config: dict) -> ConfigurationResults:
        """Returns best config and ConfigurationResults for instance set."""
        objective = self.configurator.scenario.sparkle_objective

        # Retrieve found configuration
        _, best_config = self.configurator.get_optimal_configuration(
            self.solver, instance_set, objective)

        # Retrieve validation results
        validator = Validator(self.directory)
        val_results = validator.get_validation_results(
            self.solver, instance_set, config=best_config,
            source_dir=self.directory, subdir="validation")
        header = val_results[0]
        results = []
        value_column = header.index(objective.name)
        instance_column = header.index("Instance")
        status_column = header.index("Status")
        cpu_time_column = header.index("CPU Time")
        wall_time_column = header.index("Wallclock Time")
        for res in val_results[1:]:
            results.append([res[instance_column], SolverStatus(res[status_column]),
                            res[value_column], res[cpu_time_column],
                            res[wall_time_column]])
        final_results = ValidationResults(self.solver, config,
                                          instance_set, results)
        perf_par = sgrfch.get_average_performance(val_results,
                                                  objective)
        return ConfigurationResults(perf_par,
                                    final_results)

    def serialize_configuration_results(self: ConfigurationOutput,
                                        cr: ConfigurationResults) -> dict:
        """Transform ConfigurationResults to dictionary format."""
        return {
            "performance": cr.performance,
            "results": {
                "solver": cr.results.solver.name,
                "configuration": cr.results.configuration,
                "instance_set": cr.results.instance_set.name,
                "result_header": cr.results.result_header,
                "result_values": cr.results.result_vals,
            },
        }

    def serialize_scenario(self: ConfigurationOutput,
                           scenario: ConfigurationScenario) -> dict:
        """Transform ConfigurationScenario to dictionary format."""
        return {
            "number_of_runs": scenario.number_of_runs,
            "solver_calls": scenario.solver_calls,
            "cpu_time": scenario.cpu_time,
            "wallclock_time": scenario.wallclock_time,
            "cutoff_time": scenario.cutoff_time,
            "cutoff_length": scenario.cutoff_length,
            "sparkle_objective": scenario.sparkle_objective.name,
            "use_features": scenario.use_features,
            "configurator_target": scenario.configurator_target,
            "feature_data": scenario.feature_data,
        }

    def write_output(self: ConfigurationOutput) -> None:
        """Write data into a JSON file."""
        output_data = {
            "solver": self.solver.name if self.solver else None,
            "configurator": (
                str(self.configurator.executable_path) if self.configurator else None
            ),
            "best_configuration": Solver.config_str_to_dict(self.best_config),
            "configurations": self.configurations,
            "scenario": self.serialize_scenario(self.configurator.scenario)
            if self.configurator.scenario else None,
            "training_results": [
                self.serialize_configuration_results(validation_result)
                for validation_result in self.validation_results
            ],
            "test_set": (
                [
                    self.serialize_configuration_results(validation_result)
                    for validation_result in self.validation_results_test
                ]
                if self.instance_set_test else None
            ),
        }

        self.output.parent.mkdir(parents=True, exist_ok=True)
        with self.output.open("w") as f:
            json.dump(output_data, f, indent=4)



"""Tools to parse runsolver I/O."""
import sys
from pathlib import Path
import ast
import re

from sparkle.types import SolverStatus


def get_measurements(runsolver_values_path: Path,
                     not_found: float = -1.0) -> tuple[float, float, float]:
    """Return the CPU and wallclock time reported by runsolver in values log."""
    cpu_time, wall_time, memory = not_found, not_found, not_found
    if runsolver_values_path.exists():
        with runsolver_values_path.open("r") as infile:
            lines = [line.strip().split("=") for line in infile.readlines()
                     if line.count("=") == 1]
            for keyword, value in lines:
                if keyword == "WCTIME":
                    wall_time = float(value)
                elif keyword == "CPUTIME":
                    cpu_time = float(value)
                elif keyword == "MAXVM":
                    memory = float(int(value) / 1024.0)  # MB
                    # Order is fixed, CPU is the last thing we want to read, so break
                    break
    return cpu_time, wall_time, memory


def get_status(runsolver_values_path: Path, runsolver_raw_path: Path) -> SolverStatus:
    """Get run status from runsolver logs."""
    if not runsolver_values_path.exists() and (runsolver_raw_path is not None
                                               and not runsolver_raw_path.exists()):
        # Runsolver logs were not created, job was stopped ''incorrectly''
        return SolverStatus.CRASHED
    # First check if runsolver reported time out
    if runsolver_values_path.exists():
        for line in reversed(runsolver_values_path.open("r").readlines()):
            if line.strip().startswith("TIMEOUT="):
                if line.strip() == "TIMEOUT=true":
                    return SolverStatus.TIMEOUT
                break
    if runsolver_raw_path is None:
        return SolverStatus.UNKNOWN
    if not runsolver_raw_path.exists():
        # Runsolver log was not created, job was stopped ''incorrectly''
        return SolverStatus.KILLED
    # Last line of runsolver log should contain the raw sparkle solver wrapper output
    runsolver_raw_contents = runsolver_raw_path.open("r").read().strip()
    # cutoff_time =
    sparkle_wrapper_dict_str = runsolver_raw_contents.splitlines()[-1]
    solver_regex_filter = re.findall("{.*}", sparkle_wrapper_dict_str)[0]
    output_dict = ast.literal_eval(solver_regex_filter)
    status = SolverStatus(output_dict["status"])
    # if status == SolverStatus.CRASHED and cpu_time > cutoff_time
    return status


def get_solver_args(runsolver_log_path: Path) -> str:
    """Retrieves solver arguments dict from runsolver log."""
    if runsolver_log_path.exists():
        for line in runsolver_log_path.open("r").readlines():
            if line.startswith("command line:"):
                return line.split("sparkle_solver_wrapper.py", 1)[1].strip().strip("'")
    return ""


def get_solver_output(runsolver_configuration: list[str],
                      process_output: str,
                      log_dir: Path) -> dict[str, str | object]:
    """Decode solver output dictionary when called with runsolver."""
    solver_input = None
    solver_output = None
    value_data_file = None
    cutoff_time = sys.maxsize
    for idx, conf in enumerate(runsolver_configuration):
        if not isinstance(conf, str):
            # Looking for arg names
            continue
        conf = conf.strip()
        if conf == "-o" or conf == "--solver-data":
            # solver output was redirected
            solver_data_file = Path(runsolver_configuration[idx + 1])
            if (log_dir / solver_data_file).exists():
                solver_output = (log_dir / solver_data_file).open("r").read()
        if "-v" in conf or "--var" in conf:
            value_data_file = Path(runsolver_configuration[idx + 1])
        if "--cpu-limit" in conf:
            cutoff_time = float(runsolver_configuration[idx + 1])
        if "-w" in conf or "--watcher-data" in conf:
            watch_file = Path(runsolver_configuration[idx + 1])
            args_str = get_solver_args(log_dir / watch_file)
            if args_str == "":  # Could not find log file or args
                continue
            solver_input = re.findall("{.*}", args_str)[0]
            solver_input = ast.literal_eval(solver_input)
            cutoff_time = float(solver_input["cutoff_time"])

    if solver_output is None:
        # Still empty, try to read from subprocess
        solver_output = process_output
    # Format output to only the brackets (dict)
    # NOTE: It should have only one match, do we want some error logging here?
    try:
        solver_regex_filter = re.findall("{.*}", solver_output)[0]
        output_dict = ast.literal_eval(solver_regex_filter)
    except Exception:
        config_str = " ".join(runsolver_configuration)
        print("WARNING: Solver output decoding failed from RunSolver configuration: "
              f"'{config_str}'. Setting status to 'UNKNOWN'.")
        output_dict = {"status": SolverStatus.UNKNOWN}

    output_dict["cutoff_time"] = cutoff_time
    if value_data_file is not None:
        cpu_time, wall_time, memory = get_measurements(log_dir / value_data_file)
        output_dict["cpu_time"] = cpu_time
        output_dict["wall_time"] = wall_time
        output_dict["memory"] = memory
    else:  # Could not retrieve cpu and wall time (log does not exist)
        output_dict["cpu_time"], output_dict["wall_time"] = -1.0, -1.0
    if output_dict["cpu_time"] > cutoff_time:
        output_dict["status"] = SolverStatus.TIMEOUT
    # Add the missing objectives (runtime based)
    if solver_input is not None and "objectives" in solver_input:
        objectives = solver_input["objectives"].split(",")
        for o_name in objectives:
            if o_name not in output_dict:
                output_dict[o_name] = None
    return output_dict



"""File to handle a Selector for selecting Solvers."""
from __future__ import annotations
from pathlib import Path
import subprocess
import ast

import runrunner as rrr
from runrunner import Runner, Run

from sparkle.types import SparkleCallable, SparkleObjective
from sparkle.structures import FeatureDataFrame, PerformanceDataFrame


class Selector(SparkleCallable):
    """The Selector class for handling Algorithm Selection."""

    def __init__(self: SparkleCallable,
                 executable_path: Path,
                 raw_output_directory: Path) -> None:
        """Initialize the Selector object.

        Args:
            executable_path: Path of the Selector executable.
            raw_output_directory: Directory where the Selector will write its raw output.
                Defaults to directory / tmp
        """
        self.selector_builder_path = executable_path
        self.directory = self.selector_builder_path.parent
        self.name = self.selector_builder_path.name
        self.raw_output_directory = raw_output_directory

        if not self.raw_output_directory.exists():
            self.raw_output_directory.mkdir(parents=True)

    def build_construction_cmd(
            self: Selector,
            target_file: Path,
            performance_data: Path,
            feature_data: Path,
            objective: SparkleObjective,
            runtime_cutoff: int | float | str = None,
            wallclock_limit: int | float | str = None) -> list[str | Path]:
        """Builds the commandline call string for constructing the Selector.

        Args:
            target_file: Path to the file to save the Selector to.
            performance_data: Path to the performance data csv.
            feature_data: Path to the feature data csv.
            objective: The objective to optimize for selection.
            runtime_cutoff: Cutoff for the runtime in seconds. Defaults to None
            wallclock_limit: Cutoff for total wallclock in seconds. Defaults to None

        Returns:
            The command list for constructing the Selector.
        """
        objective_function = "runtime" if objective.time else "solution_quality"
        # Python3 to avoid execution rights
        cmd = ["python3", self.selector_builder_path,
               "--performance_csv", performance_data,
               "--feature_csv", feature_data,
               "--objective", objective_function,
               "--save", target_file]
        if runtime_cutoff is not None:
            cmd.extend(["--runtime_cutoff", str(runtime_cutoff), "--tune"])
        if wallclock_limit is not None:
            cmd.extend(["--wallclock_limit", str(wallclock_limit)])
        return cmd

    def construct(self: Selector,
                  target_file: Path | str,
                  performance_data: PerformanceDataFrame,
                  feature_data: FeatureDataFrame,
                  objective: SparkleObjective,
                  runtime_cutoff: int | float | str = None,
                  wallclock_limit: int | float | str = None,
                  run_on: Runner = Runner.SLURM,
                  sbatch_options: list[str] = None,
                  base_dir: Path = Path()) -> Run:
        """Construct the Selector.

        Args:
            target_file: Path to the file to save the Selector to.
            performance_data: Path to the performance data csv.
            feature_data: Path to the feature data csv.
            objective: The objective to optimize for selection.
            runtime_cutoff: Cutoff for the runtime in seconds.
            wallclock_limit: Cutoff for the wallclock time in seconds.
            run_on: Which runner to use. Defaults to slurm.
            sbatch_options: Additional options to pass to sbatch.
            base_dir: The base directory to run the Selector in.

        Returns:
            Path to the constructed Selector.
        """
        if isinstance(target_file, str):
            target_file = self.raw_output_directory / target_file
        # Convert the dataframes to Selector Format
        performance_csv = performance_data.to_autofolio(objective=objective,
                                                        target=target_file.parent)
        feature_csv = feature_data.to_autofolio(target_file.parent)
        cmd = self.build_construction_cmd(target_file,
                                          performance_csv,
                                          feature_csv,
                                          objective,
                                          runtime_cutoff,
                                          wallclock_limit)

        cmd_str = " ".join([str(c) for c in cmd])
        construct = rrr.add_to_queue(
            runner=run_on,
            cmd=[cmd_str],
            name="construct_selector",
            base_dir=base_dir,
            stdout=Path("normal.log"),
            stderr=Path("error.log"),
            sbatch_options=sbatch_options)
        if run_on == Runner.LOCAL:
            construct.wait()
            if not target_file.is_file():
                print(f"Selector construction of {self.name} failed!")

        return construct

    def build_cmd(self: Selector,
                  selector_path: Path,
                  feature_vector: list | str) -> list[str | Path]:
        """Builds the commandline call string for running the Selector."""
        if isinstance(feature_vector, list):
            feature_vector = " ".join(map(str, feature_vector))

        return ["python3", self.selector_builder_path,
                "--load", selector_path,
                "--feature_vec", feature_vector]

    def run(self: Selector,
            selector_path: Path,
            feature_vector: list | str) -> list:
        """Run the Selector, returning the prediction schedule upon success."""
        cmd = self.build_cmd(selector_path, feature_vector)
        run = subprocess.run(cmd, capture_output=True)
        if run.returncode != 0:
            print(f"Selector run of {self.name} failed! Error:\n"
                  f"{run.stderr.decode()}")
            return None
        # Process the prediction schedule from the output
        schedule = Selector.process_predict_schedule_output(run.stdout.decode())
        if schedule is None:
            print(f"Error getting predict schedule! Selector {self.name} output:\n"
                  f"{run.stderr.decode()}")
        return schedule

    @staticmethod
    def process_predict_schedule_output(output: str) -> list:
        """Return the predicted algorithm schedule as a list."""
        prefix_string = "Selected Schedule [(algorithm, budget)]: "
        predict_schedule = ""
        predict_schedule_lines = output.splitlines()
        for line in predict_schedule_lines:
            if line.strip().startswith(prefix_string):
                predict_schedule = line.strip()
                break
        if predict_schedule == "":
            return None
        predict_schedule_string = predict_schedule[len(prefix_string):]
        return ast.literal_eval(predict_schedule_string)



#!/usr/bin/env python3
"""Sparkle command to remove an instance set from the Sparkle platform."""

import sys
import argparse
import shutil

from sparkle.CLI.help import global_variables as gv
from sparkle.platform import file_help as sfh
from sparkle.structures import FeatureDataFrame, PerformanceDataFrame
from sparkle.instance import instance_set
from sparkle.CLI.help import logging as sl
from sparkle.platform import CommandName, COMMAND_DEPENDENCIES
from sparkle.CLI.initialise import check_for_initialise
from sparkle.CLI.help import argparse_custom as ac
from sparkle.CLI.help.nicknames import resolve_object_name


def parser_function() -> argparse.ArgumentParser:
    """Define the command line arguments."""
    parser = argparse.ArgumentParser()
    parser.add_argument(*ac.InstancesPathRemoveArgument.names,
                        **ac.InstancesPathRemoveArgument.kwargs)
    return parser


if __name__ == "__main__":
    # Log command call
    sl.log_command(sys.argv)

    # Define command line arguments
    parser = parser_function()

    # Process command line arguments
    args = parser.parse_args()
    instances_path = resolve_object_name(args.instances_path,
                                         target_dir=gv.settings().DEFAULT_instance_dir)

    check_for_initialise(COMMAND_DEPENDENCIES[CommandName.REMOVE_INSTANCES])

    if instances_path is None or not instances_path.exists() or not\
            instances_path.is_dir():
        print(f'Could not resolve instances path arg "{args.instances_path}"!')
        print("Check that the path or nickname is spelled correctly.")
        sys.exit(-1)

    print(f"Start removing all instances in directory {instances_path} ...")
    old_instance_set = instance_set(instances_path)
    # Remove from feature data and performance data
    feature_data = FeatureDataFrame(gv.settings().DEFAULT_feature_data_path)
    performance_data = PerformanceDataFrame(gv.settings().DEFAULT_performance_data_path)
    for instance in old_instance_set.instance_paths:
        feature_data.remove_instances(str(instance))
        performance_data.remove_instance(str(instance))

    feature_data.save_csv()
    performance_data.save_csv()

    # Remove nickname, if it exists
    instances_nicknames = gv.file_storage_data_mapping[gv.instances_nickname_path]
    for key in instances_nicknames:
        if instances_nicknames[key] == instances_path:
            sfh.add_remove_platform_item(instances_path,
                                         gv.instances_nickname_path,
                                         key=key, remove=True)
            break

    # Remove the directory and all its files
    shutil.rmtree(instances_path)

    print(f"Removing instances in directory {instances_path} done!")



#!/usr/bin/env python3
# -*- coding: UTF-8 -*-
"""Module to manage performance data files and common operations on them."""
from __future__ import annotations
from pathlib import Path
import sys
import math
import pandas as pd

from sparkle.types import SparkleObjective, resolve_objective


class PerformanceDataFrame():
    """Class to manage performance data and common operations on them."""
    missing_value = math.nan
    missing_objective = "UNKNOWN"
    multi_dim_names = ["Objective", "Instance", "Run"]

    def __init__(self: PerformanceDataFrame,
                 csv_filepath: Path,
                 solvers: list[str] = [],
                 objectives: list[str | SparkleObjective] = None,
                 instances: list[str] = [],
                 n_runs: int = 1,
                 init_df: bool = True) -> None:
        """Initialise a PerformanceDataFrame.

        Consists of:
            - Columns representing the Solvers
            - Rows representing the result by multi-index in order of:
                * Objective (Static, given in constructor or read from file)
                * Instance
                * Runs (Static, given in constructor or read from file)

        Args:
            csv_filepath: If path exists, load from Path.
                Otherwise create new and save to this path.
            solvers: List of solver names to be added into the Dataframe
            objectives: List of SparkleObjectives or objective names. By default None,
                then the objectives will be derived from Sparkle Settings if possible.
            instances: List of instance names to be added into the Dataframe
            n_runs: The number of runs to consider per Solver/Objective/Instance comb.
            init_df: Whether the dataframe should be initialised. Set to false to reduce
                heavy IO loads.
        """
        self.csv_filepath = csv_filepath
        # Runs is a ``static'' dimension
        self.n_runs = n_runs
        self.run_ids = list(range(1, self.n_runs + 1))  # We count runs from 1
        if objectives is not None:
            self.objectives = [resolve_objective(o) if isinstance(o, str) else o
                               for o in objectives]
        else:
            self.objectives = [SparkleObjective(PerformanceDataFrame.missing_objective)]
        if init_df:
            if self.csv_filepath.exists():
                self.dataframe = pd.read_csv(csv_filepath)
                has_rows = len(self.dataframe.index) > 0
                if (PerformanceDataFrame.multi_dim_names[0] not in self.dataframe.columns
                        or not has_rows):
                    # No objective present, force into column
                    if objectives is None:
                        self.dataframe[PerformanceDataFrame.multi_dim_names[0]] =\
                            PerformanceDataFrame.missing_objective
                    else:  # Constructor is provided with the objectives
                        self.dataframe[PerformanceDataFrame.multi_dim_names[0]] =\
                            [o.name for o in self.objectives]
                else:
                    # Objectives are present, determine which ones
                    names = self.dataframe[PerformanceDataFrame.multi_dim_names[0]]
                    self.objectives = [resolve_objective(name) for name in
                                       names.unique()]
                if (PerformanceDataFrame.multi_dim_names[2] not in self.dataframe.columns
                        or not has_rows):
                    # No runs column present, force into column
                    self.n_runs = 1
                    self.dataframe[PerformanceDataFrame.multi_dim_names[2]] = self.n_runs
                    self.run_ids = [self.n_runs]
                else:
                    # Runs are present, determine run ids
                    run_label = PerformanceDataFrame.multi_dim_names[2]
                    self.run_ids = self.dataframe[run_label].unique().tolist()
                if PerformanceDataFrame.multi_dim_names[1] not in self.dataframe.columns:
                    # Instances are listed as rows, force into column
                    self.dataframe = self.dataframe.reset_index().rename(
                        columns={"index": PerformanceDataFrame.multi_dim_names[1]})
                # Now we can cast the columns into multi dim
                self.dataframe = self.dataframe.set_index(
                    PerformanceDataFrame.multi_dim_names)
            else:
                # Initialize empty DataFrame
                midx = pd.MultiIndex.from_product(
                    [[o.name for o in self.objectives], instances, self.run_ids],
                    names=PerformanceDataFrame.multi_dim_names)
                self.dataframe = pd.DataFrame(PerformanceDataFrame.missing_value,
                                              index=midx,
                                              columns=solvers)
                self.save_csv()
            # Sort the index to optimize lookup speed
            self.dataframe = self.dataframe.sort_index()

    def __repr__(self: PerformanceDataFrame) -> str:
        """Return string representation of the DataFrame."""
        return self.dataframe.__repr__()

    # Properties

    @property
    def num_objectives(self: PerformanceDataFrame) -> int:
        """Retrieve the number of objectives in the DataFrame."""
        return self.dataframe.index.levels[0].size

    @property
    def num_instances(self: PerformanceDataFrame) -> int:
        """Return the number of instances."""
        return self.dataframe.index.levels[1].size

    @property
    def num_runs(self: PerformanceDataFrame) -> int:
        """Return the number of runs."""
        return self.dataframe.index.levels[2].size

    @property
    def num_solvers(self: PerformanceDataFrame) -> int:
        """Return the number of solvers."""
        return self.dataframe.columns.size

    @property
    def multi_objective(self: PerformanceDataFrame) -> bool:
        """Return whether the dataframe represent MO or not."""
        return self.num_objectives > 1

    @property
    def solvers(self: PerformanceDataFrame) -> list[str]:
        """Return the solver present as a list of strings."""
        return self.dataframe.columns.tolist()

    @property
    def objective_names(self: PerformanceDataFrame) -> list[str]:
        """Return the objective names as a list of strings."""
        if self.num_objectives == 0:
            return [PerformanceDataFrame.missing_objective]
        return self.dataframe.index.levels[0].tolist()

    @property
    def instances(self: PerformanceDataFrame) -> list[str]:
        """Return the instances as a Pandas Index object."""
        return self.dataframe.index.levels[1].tolist()

    @property
    def has_missing_values(self: PerformanceDataFrame) -> bool:
        """Returns True if there are any missing values in the dataframe."""
        return self.dataframe.isnull().any().any()

    def verify_objective(self: PerformanceDataFrame,
                         objective: str) -> str:
        """Method to check whether the specified objective is valid.

        Users are allowed to index the dataframe without specifying all dimensions.
        However, when dealing with multiple objectives this is not allowed and this
        is verified here. If we have only one objective this is returned. Otherwise,
        if an objective is specified by the user this is returned.

        Args:
            objective: The objective given by the user
        """
        if objective is None:
            if self.multi_objective:
                raise ValueError("Error: MO Data, but objective not specified.")
            elif self.num_objectives == 1:
                return self.objective_names[0]
            else:
                return PerformanceDataFrame.missing_objective
        return objective

    def verify_run_id(self: PerformanceDataFrame,
                      run_id: int) -> int:
        """Method to check whether run id is valid.

        Similar to verify_objective but here we check the dimensionality of runs.

        Args:
            run_id: the run as specified by the user.
        """
        if run_id is None:
            if self.n_runs > 1:
                print("Error: Multiple run performance data, but run not specified")
                sys.exit(-1)
            else:
                run_id = self.run_ids[0]
        return run_id

    def verify_indexing(self: PerformanceDataFrame,
                        objective: str,
                        run_id: int) -> tuple[str, int]:
        """Method to check whether data indexing is correct.

        Users are allowed to use the Performance Dataframe without the second and
        fourth dimension (Objective and Run respectively) in the case they only
        have one objective or only do one run. This method adjusts the indexing for
        those cases accordingly.

        Args:
            objective: The given objective name
            run_id: The given run index

        Returns:
            A tuple representing the (possibly adjusted) Objective and Run index.
        """
        objective = self.verify_objective(objective)
        run_id = self.verify_run_id(run_id)
        return objective, run_id

    # Getters and Setters

    def add_solver(self: PerformanceDataFrame,
                   solver_name: str,
                   initial_value: float | list[float] = None) -> None:
        """Add a new solver to the dataframe. Initializes value to None by default.

        Args:
            solver_name: The name of the solver to be added.
            initial_value: The value assigned for each index of the new solver.
                If not None, must match the index dimension (n_obj * n_inst * n_runs).
        """
        if solver_name in self.dataframe.columns:
            print(f"WARNING: Tried adding already existing solver {solver_name} to "
                  f"Performance DataFrame: {self.csv_filepath}")
            return
        self.dataframe[solver_name] = initial_value

    def add_instance(self: PerformanceDataFrame,
                     instance_name: str,
                     initial_value: float | list[float] = None) -> None:
        """Add and instance to the DataFrame."""
        if self.dataframe.index.size == 0 or self.dataframe.columns.size == 0:
            # First instance or no Solvers yet
            solvers = self.dataframe.columns.to_list()
            instances = self.dataframe.index.levels[1].to_list() + [instance_name]
            midx = pd.MultiIndex.from_product(
                [self.objective_names, instances, self.run_ids],
                names=PerformanceDataFrame.multi_dim_names)
            self.dataframe = pd.DataFrame(initial_value, index=midx, columns=solvers)
        else:
            if instance_name in self.dataframe.index.levels[1]:
                print(f"WARNING: Tried adding already existing instance {instance_name} "
                      f"to Performance DataFrame: {self.csv_filepath}")
                return
            # Create the missing indices
            levels = [self.dataframe.index.levels[0].tolist(),
                      [instance_name],
                      self.dataframe.index.levels[2].tolist()]
            # NOTE: Did this fix Jeroen's bug? .from_arrays instead of direct constructor
            emidx = pd.MultiIndex.from_arrays(levels,
                                              names=PerformanceDataFrame.multi_dim_names)
            # Create the missing column values
            edf = pd.DataFrame(PerformanceDataFrame.missing_value,
                               index=emidx,
                               columns=self.dataframe.columns)
            # Concatenate the original and new dataframe together
            self.dataframe = pd.concat([self.dataframe, edf])

    # Can we make this handle a sequence of inputs instead of just 1?
    def set_value(self: PerformanceDataFrame,
                  value: float,
                  solver: str,
                  instance: str,
                  objective: str = None,
                  run: int = None) -> None:
        """Setter method to assign a value to the Dataframe.

        Args:
            value: Float value to be assigned.
            solver: The solver that produced the value.
            instance: The instance that the value was produced on.
            objective: The objective for which the result was produced.
                Optional in case of using single objective.
            run: The run index for which the result was produced.
                Optional in case of doing single run results.
        """
        objective, run = self.verify_indexing(objective, run)
        self.dataframe.at[(objective, instance, run), solver] = value

    def remove_solver(self: PerformanceDataFrame, solver_name: str | list[str]) -> None:
        """Drop one or more solvers from the Dataframe."""
        self.dataframe.drop(solver_name, axis=1, inplace=True)

    def remove_instance(self: PerformanceDataFrame, instance_name: str) -> None:
        """Drop an instance from the Dataframe."""
        self.dataframe.drop(instance_name, axis=0, level="Instance", inplace=True)

    def reset_value(self: PerformanceDataFrame,
                    solver: str,
                    instance: str,
                    objective: str = None,
                    run: int = None) -> None:
        """Reset a value in the dataframe."""
        self.set_value(PerformanceDataFrame.missing_value,
                       solver, instance, objective, run)

    # Can we unify get_value and get_values?
    def get_value(self: PerformanceDataFrame,
                  solver: str,
                  instance: str,
                  objective: str = None,
                  run: int = None) -> float:
        """Index a value of the DataFrame and return it."""
        objective, run = self.verify_indexing(objective, run)
        return float(self.dataframe.loc[(objective, instance, run), solver])

    def get_values(self: PerformanceDataFrame,
                   solver: str,
                   instance: str = None,
                   objective: str = None,
                   run: int = None) -> list[float]:
        """Return a list of solver values."""
        subdf = self.dataframe[solver]
        if objective is not None:
            objective = self.verify_objective(objective)
            subdf = subdf.xs(objective, level=0, drop_level=False)
        if instance is not None:
            subdf = subdf.xs(instance, level=1, drop_level=False)
        if run is not None:
            run = self.verify_run_id(run)
            subdf = subdf.xs(run, level=2, drop_level=False)
        return subdf.to_list()

    # Modifiers

    '''def penalise(self: PerformanceDataFrame,
                 threshold: float,
                 penalty: float,
                 objective: str = None,
                 lower_bound: bool = False) -> None:
        """Penalises the DataFrame values if crossing threshold by specified penalty.

        Directly updates the DataFrame object held by this class.

        Args:
            threshold: The threshold of performances to be met
            penalty: The values assigned for out of bounds performances
            objective: The objective that should be penalised.
            lower_bound: Whether the threshold is a lower_bound. By default,
                the threshold is treated as an upperbound for performance values.
        """
        objective = self.verify_objective(objective)
        comparison_op = operator.lt if lower_bound else operator.gt
        self.dataframe[comparison_op(self.dataframe.loc[(objective), :],
                                     threshold)] = penalty'''

    # Calculables

    def mean(self: PerformanceDataFrame,
             objective: str = None,
             solver: str = None,
             instance: str = None) -> float:
        """Return the mean value of a slice of the dataframe."""
        objective = self.verify_objective(objective)
        subset = self.dataframe.xs(objective, level=0)
        if solver is not None:
            subset = subset.xs(solver, axis=1, drop_level=False)
        if instance is not None:
            subset = subset.xs(instance, axis=0, drop_level=False)
        value = subset.astype(float).mean()
        if isinstance(value, pd.Series):
            return value.mean()
        return value

    # TODO: This method should be refactored or not exist
    def get_job_list(self: PerformanceDataFrame, rerun: bool = False) \
            -> list[tuple[str, str]]:
        """Return a list of performance computation jobs there are to be done.

        Get a list of tuple[instance, solver] to run from the performance data
        csv file. If rerun is False (default), get only the tuples that don't have a
        value in the table, else (True) get all the tuples.

        Args:
            rerun: Boolean indicating if we want to rerun all jobs
        """
        df = self.dataframe.stack(future_stack=True)
        if not rerun:
            df = df[df.isnull()]
        df.index = df.index.droplevel(["Objective"])
        return df.index.unique().tolist()

    # TODO: This method should be refactored or not exist
    def remaining_jobs(self: PerformanceDataFrame) -> dict[str, list[str]]:
        """Return a dictionary for empty values per instance and solver combination."""
        remaining_jobs = {}
        null_df = self.dataframe.isnull()
        for row in self.dataframe.index:
            instance = row[1]
            for solver in self.dataframe.columns:
                if null_df.at[row, solver]:
                    if instance not in remaining_jobs:
                        remaining_jobs[instance] = set([solver])
                    else:
                        remaining_jobs[instance].add(solver)
        return remaining_jobs

    def best_instance_performance(
            self: PerformanceDataFrame,
            objective: str | SparkleObjective = None,
            run_id: int = None,
            exclude_solvers: list[str] = None) -> pd.Series:
        """Return the best performance for each instance in the portfolio.

        Args:
            objective: The objective for which we calculate the best performance
            run_id: The run for which we calculate the best performance. If None,
                we consider all runs.
            exclude_solvers: List of solvers to exclude in the calculation.

        Returns:
            The best performance for each instance in the portfolio.
        """
        objective = self.verify_objective(objective)
        if isinstance(objective, str):
            objective = resolve_objective(objective)
        subdf = self.dataframe.xs(objective.name, level=0)
        if exclude_solvers is not None:
            subdf = subdf.drop(exclude_solvers, axis=1)
        if run_id is not None:
            run_id = self.verify_run_id(run_id)
            subdf = subdf.xs(run_id, level=1)
        else:
            # Drop the run level
            subdf = subdf.droplevel(level=1)
        if objective.minimise:
            series = subdf.min(axis=1)
        else:
            series = subdf.max(axis=1)
        # Ensure we always return the best for each run
        series = series.sort_values(ascending=objective.minimise)
        return series.groupby(series.index).first().astype(float)

    def best_performance(
            self: PerformanceDataFrame,
            exclude_solvers: list[str] = [],
            objective: str | SparkleObjective = None) -> float:
        """Return the overall best performance of the portfolio.

        Args:
            exclude_solvers: List of solvers to exclude in the calculation.
                Defaults to none.
            objective: The objective for which we calculate the best performance

        Returns:
            The aggregated best performance of the portfolio over all instances.
        """
        objective = self.verify_objective(objective)
        if isinstance(objective, str):
            objective = resolve_objective(objective)
        instance_best = self.best_instance_performance(
            objective, exclude_solvers=exclude_solvers).to_numpy(dtype=float)
        return objective.instance_aggregator(instance_best)

    def schedule_performance(
            self: PerformanceDataFrame,
            schedule: dict[str: list[tuple[str, float | None]]],
            target_solver: str = None,
            objective: str | SparkleObjective = None) -> float:
        """Return the performance of a selection schedule on the portfolio.

        Args:
            schedule: Compute the best performance according to a selection schedule.
                A dictionary with instances as keys and a list of tuple consisting of
                (solver, max_runtime) or solvers if no runtime prediction should be used.
            target_solver: If not None, store the values in this solver of the DF.
            objective: The objective for which we calculate the best performance

        Returns:
            The performance of the schedule over the instances in the dictionary.
        """
        objective = self.verify_objective(objective)
        if isinstance(objective, str):
            objective = resolve_objective(objective)
        select = min if objective.minimise else max
        performances = [0.0 for _ in range(len(schedule.keys()))]
        for ix, instance in enumerate(schedule.keys()):
            for iy, (solver, max_runtime) in enumerate(schedule[instance]):
                performance = self.get_value(solver, instance, objective.name)
                if max_runtime is not None:  # We are dealing with runtime
                    performances[ix] += performance
                    if performance < max_runtime:
                        break  # Solver finished in time
                else:  # Quality, we take the best found performance
                    if iy == 0:  # First solver, set initial value
                        performances[ix] = performance
                        continue
                    performances[ix] = select(performances[ix], performance)
            if target_solver is not None:
                self.set_value(performances[ix], target_solver, instance, objective.name)
        return performances

    def marginal_contribution(
            self: PerformanceDataFrame,
            objective: str | SparkleObjective = None,
            sort: bool = False) -> list[float]:
        """Return the marginal contribution of the solvers on the instances.

        Args:
            objective: The objective for which we calculate the marginal contribution.
            sort: Whether to sort the results afterwards
        Returns:
            The marginal contribution of each solver.
        """
        output = []
        objective = self.verify_objective(objective)
        if isinstance(objective, str):
            objective = resolve_objective(objective)
        best_performance = self.best_performance(objective=objective)
        for solver in self.solvers:
            # By calculating the best performance excluding this Solver,
            # we can determine its relative impact on the portfolio.
            missing_solver_best = self.best_performance(
                exclude_solvers=[solver],
                objective=objective)
            # Now we need to see how much the portfolio's best performance
            # decreases without this solver.
            marginal_contribution = missing_solver_best / best_performance
            if missing_solver_best == best_performance:
                # No change, no contribution
                marginal_contribution = 0.0
            output.append((solver, marginal_contribution, missing_solver_best))
        if sort:
            output.sort(key=lambda x: x[1], reverse=objective.minimise)
        return output

    def get_solver_ranking(self: PerformanceDataFrame,
                           objective: str | SparkleObjective = None
                           ) -> list[tuple[str, float]]:
        """Return a list with solvers ranked by average performance."""
        objective = self.verify_objective(objective)
        if isinstance(objective, str):
            objective = resolve_objective(objective)
        sub_df = self.dataframe.loc(axis=0)[objective.name, :, :]
        # Reduce Runs Dimension
        sub_df = sub_df.droplevel("Run")
        sub_df = sub_df.groupby(sub_df.index).agg(func=objective.run_aggregator)
        solver_ranking = [(solver, objective.instance_aggregator(
            sub_df[solver].astype(float))) for solver in self.solvers]
        # Sort the list by second value (the performance)
        solver_ranking.sort(key=lambda performance: performance[1],
                            reverse=(not objective.minimise))
        return solver_ranking

    def save_csv(self: PerformanceDataFrame, csv_filepath: Path = None) -> None:
        """Write a CSV to the given path.

        Args:
            csv_filepath: String path to the csv file. Defaults to self.csv_filepath.
        """
        csv_filepath = self.csv_filepath if csv_filepath is None else csv_filepath
        self.dataframe.to_csv(csv_filepath)

    def clean_csv(self: PerformanceDataFrame) -> None:
        """Set all values in Performance Data to None."""
        self.dataframe[:] = PerformanceDataFrame.missing_value
        self.save_csv()

    def copy(self: PerformanceDataFrame,
             csv_filepath: Path = None) -> PerformanceDataFrame:
        """Create a copy of this object.

        Args:
            csv_filepath: The new filepath to use for saving the object to.
                Warning: If the original path is used, it could lead to dataloss!
        """
        csv_filepath = self.csv_filepath if csv_filepath is None else csv_filepath
        pd_copy = PerformanceDataFrame(self.csv_filepath, init_df=False)
        pd_copy.dataframe = self.dataframe.copy()
        pd_copy.csv_filepath = csv_filepath
        return pd_copy

    def to_autofolio(self: PerformanceDataFrame,
                     objective: SparkleObjective = None,
                     target: Path = None) -> Path:
        """Port the data to a format acceptable for AutoFolio."""
        if (objective is None and self.multi_objective or self.n_runs > 1):
            print(f"ERROR: Currently no porting available for {self.csv_filepath} "
                  "to Autofolio due to multi objective or number of runs.")
            return
        autofolio_df = self.dataframe.copy()
        if objective is not None:
            autofolio_df = autofolio_df.loc[objective.name]
            autofolio_df.index = autofolio_df.index.droplevel("Run")
        else:
            autofolio_df.index = autofolio_df.index.droplevel(["Objective", "Run"])
        if target is None:
            path = self.csv_filepath.parent / f"autofolio_{self.csv_filepath.name}"
        else:
            path = target / f"autofolio_{self.csv_filepath.name}"
        autofolio_df.to_csv(path)
        return path



"""Tests for feature dataframe class."""


def test_feature_dataframe_constructor() -> None:
    """Test feature dataframe constructor."""
    # TODO: write test
    pass


def test_add_extractor() -> None:
    """Test for method add_extractor."""
    # TODO: Write test
    pass


def test_add_instances() -> None:
    """Test for method add_instances."""
    # TODO: Write test
    pass


def test_remove_extractor() -> None:
    """Test for method remove_extractor."""
    # TODO: Write test
    pass


def test_remove_instances() -> None:
    """Test for method remove_instances."""
    # TODO: Write test
    pass


def test_get_feature_groups() -> None:
    """Test for method get_feature_groups."""
    # TODO: Write test
    pass


def get_value() -> None:
    """Test for method get_value."""
    # TODO: Write test
    pass


def set_value() -> None:
    """Test for method set_value."""
    # TODO: Write test
    pass


def test_has_missing_vectors() -> None:
    """Test for method has_missing_vectors."""
    # TODO: Write test for completely filled FDF
    # TODO: Write test for:
    #  - atleast one missing value per instance for all extractors
    #  - One feature group being compeletely missing
    #  - One instance being completely missing
    pass


def test_get_remaining_jobs() -> None:
    """Test for method get_remaining_jobs."""
    # TODO: Write test
    pass


def test_get_instance() -> None:
    """Test for method get_instance."""
    # TODO: Write test
    pass


def test_impute_missing_values() -> None:
    """Test for method impute_missing_values."""
    # TODO: Write test for one missing value (Can be calculated)
    # TODO: Write test for one missing feature in every instance
    pass


def test_has_missing_values() -> None:
    """Test for method has_missing_values."""
    # TODO: Write test
    pass


def test_reset_dataframe() -> None:
    """Test for method reset_dataframe."""
    # TODO: Write test
    pass


def test_sort() -> None:
    """Test for method sort."""
    # TODO: Write test
    pass


def test_instances() -> None:
    """Test instances property."""
    # TODO: Write test
    pass


def test_extractors() -> None:
    """Test for property extractors."""
    # TODO: Write test
    pass


def test_save_csv() -> None:
    """Test for method save_csv."""
    # TODO: Write test
    pass


def test_to_autofolio() -> None:
    """Test for method to_autofolio."""
    # TODO: Write test
    pass



"""Tests for all helper functions in sparkle_generate_report_for_configuration_help."""

import pytest
from pathlib import Path

from pytest_mock import MockFixture

from sparkle.solver.ablation import AblationScenario
from sparkle.platform import generate_report_for_configuration as sgrch
from sparkle.platform.settings_objects import Settings
from sparkle.solver.validator import Validator
from sparkle.types.objective import SparkleObjective, PAR
from sparkle.solver import Solver
from sparkle.instance import instance_set
import csv

global settings
settings = Settings()

configurator = settings.get_general_sparkle_configurator()
solver_path = Path("tests/test_files/Solvers/Test-Solver")
solver = Solver(solver_path, raw_output_directory=Path(""))
train_instance = "train-instance"
test_instance = "test-instance"
configurator_path = configurator.configurator_path
configurator.scenario =\
    configurator.scenario_class(solver, Path(train_instance))
configurator.scenario._set_paths(configurator_path)
ablation_scenario = AblationScenario(
    solver, Path(train_instance), Path(test_instance), Path(""))
validator = Validator()
test_objective_runtime = PAR(10)
test_objective_quality = SparkleObjective("ACCURACY")


def test_get_average_performance(mocker: MockFixture) -> None:
    """Test get_average_performance returns correct value.

    A performance list should be retrieved from results file.
    The mean of the performance values should be computed and returned.
    """
    mock_get_list = mocker.patch("sparkle.platform.generate_report_for_configuration."
                                 "get_dict_instance_to_performance",
                                 return_value={"one": 10, "two": 5})

    avg = sgrch.get_average_performance([], test_objective_runtime)
    mock_get_list.assert_called_once_with([], test_objective_runtime)
    assert avg == 7.5


def test_get_dict_instance_to_performance(mocker: MockFixture) -> None:
    """Test get_dict_instance_to_performance creates dict from performance list."""
    validation_file = Path("tests/test_files/Validator/validation.csv")
    csv_data = [line for line in csv.reader(validation_file.open("r"))]
    instance_dict = sgrch.get_dict_instance_to_performance(
        csv_data, test_objective_runtime)
    assert instance_dict == {
        "bce7824.cnf": 10.1316,
        "Ptn-7824-b09.cnf": 10.0667,
        "Ptn-7824-b13.cnf": 10.0786,
        "Ptn-7824-b21.cnf": 10.1152,
        "Ptn-7824-b15.cnf": 10.1177,
        "Ptn-7824-b01.cnf": 10.0812,
        "Ptn-7824-b19.cnf": 10.0642,
        "Ptn-7824-b05.cnf": 10.1184,
        "Ptn-7824-b03.cnf": 10.1368,
        "Ptn-7824-b11.cnf": 10.1921,
        "Ptn-7824-b07.cnf": 10.1672,
        "Ptn-7824-b17.cnf": 0.993013
    }


def test_get_ablation_bool_true(mocker: MockFixture) -> None:
    """Test get_ablation_bool returns correct string if get_ablation_bool is True."""
    mock_check = mocker.patch("sparkle.solver.ablation.AblationScenario."
                              "check_for_ablation",
                              return_value=True)
    ablation_bool = sgrch.get_ablation_bool(ablation_scenario)
    mock_check.assert_called_once()
    assert ablation_bool == r"\ablationtrue"


def test_get_ablation_bool_false(mocker: MockFixture) -> None:
    """Test get_ablation_bool returns correct string if get_ablation_bool is False."""
    mock_check = mocker.patch("sparkle.solver.ablation.AblationScenario."
                              "check_for_ablation",
                              return_value=False)
    ablation_bool = sgrch.get_ablation_bool(ablation_scenario)
    mock_check.assert_called_once()
    assert ablation_bool == r"\ablationfalse"


def test_get_data_for_plot_same_instance(mocker: MockFixture) -> None:
    """Test get_data_for_plot returns list of values if dicts are correct."""
    dict_configured = {"instance-1.cnf": 1.0}
    dict_default = {"instance-1.cnf": 0.01}
    mock_dict = mocker.patch("sparkle.platform.generate_report_for_configuration."
                             "get_dict_instance_to_performance",
                             side_effect=[dict_configured, dict_default])

    configured_dir = "configured/directory/"
    default_dir = "default/directory/"
    points = sgrch.get_data_for_plot(
        configured_dir, default_dir, test_objective_runtime)

    mock_dict.assert_any_call(default_dir, test_objective_runtime)
    mock_dict.assert_any_call(configured_dir, test_objective_runtime)
    assert points == [[1.0, 0.01]]


def test_get_data_for_plot_instance_error(mocker: MockFixture) -> None:
    """Test get_data_for_plot raises a SystemExit if dicts to not fit.

    If the two dicts do not contain the same instances, an error is raised.
    """
    dict_configured = {
        "instance-2.cnf": 1.0
    }
    dict_default = {
        "instance-1.cnf": 0.01
    }
    mock_dict = mocker.patch("sparkle.platform.generate_report_for_configuration."
                             "get_dict_instance_to_performance",
                             side_effect=[dict_configured, dict_default])

    configured_dir = "configured/directory/"
    default_dir = "default/directory/"
    with pytest.raises(SystemExit):
        sgrch.get_data_for_plot(
            configured_dir, default_dir, test_objective_runtime)

    mock_dict.assert_any_call(default_dir, test_objective_runtime)
    mock_dict.assert_any_call(default_dir, test_objective_runtime)


def test_get_figure_configure_vs_default(mocker: MockFixture) -> None:
    """Test get_figure_configure_vs_default creates plot and returns correct string.

    The function `generate_comparison_plot()` should be called with the correct
    arguments.
    Also, the correct LaTeX string to include the figure should be returned.
    """
    configured_dir = "configured/directory/"
    default_dir = "default/directory/"
    reports_dir = Path("configuration/report")
    filename = "figure.jpg"
    cutoff = 0
    points = [[1.0, 0.1]]
    performance_measure = "PERF_MEASURE"
    plot_params = {"xlabel": f"Default parameters [{performance_measure}]",
                   "ylabel": f"Configured parameters [{performance_measure}]",
                   "output_dir": reports_dir,
                   "scale": "linear",
                   "limit_min": 1.5,
                   "limit_max": 1.5,
                   "limit": "relative",
                   "replace_zeros": False,
                   }
    mock_data = mocker.patch("sparkle.platform.generate_report_for_configuration."
                             "get_data_for_plot",
                             return_value=points)
    mock_plot = mocker.patch("sparkle.platform.latex.generate_comparison_plot")

    figure_string = sgrch.get_figure_configure_vs_default(configured_dir,
                                                          default_dir,
                                                          reports_dir,
                                                          filename,
                                                          performance_measure,
                                                          cutoff,
                                                          test_objective_runtime)

    mock_data.assert_called_once_with(configured_dir, default_dir,
                                      test_objective_runtime)
    mock_plot.assert_called_once_with(points, filename, **plot_params)
    assert figure_string == f"\\includegraphics[width=0.6\\textwidth]{{{filename}}}"


def test_get_figure_configure_vs_default_par(mocker: MockFixture) -> None:
    """Test get_figure_configure_vs_default adds params for performance measure PAR.

    If the performance measure starts with PAR, `generate_comparison_plot()` should
    be called with additional parameters.
    """
    configured_dir = "configured/directory/"
    default_dir = "default/directory/"
    reports_dir = Path("configuration/report")
    filename = "figure.jpg"
    cutoff = 1

    points = [[1.0, 0.1]]
    performance_measure = "PAR12"
    plot_params = {"xlabel": f"Default parameters [{performance_measure}]",
                   "ylabel": f"Configured parameters [{performance_measure}]",
                   "scale": "log",
                   "limit_min": 0.25,
                   "limit_max": 0.25,
                   "limit": "magnitude",
                   "replace_zeros": True,
                   "output_dir": reports_dir
                   }
    mock_data = mocker.patch("sparkle.platform.generate_report_for_configuration."
                             "get_data_for_plot",
                             return_value=points)
    mock_plot = mocker.patch("sparkle.platform.latex.generate_comparison_plot")

    figure_string = sgrch.get_figure_configure_vs_default(configured_dir,
                                                          default_dir,
                                                          reports_dir,
                                                          filename,
                                                          performance_measure,
                                                          cutoff,
                                                          test_objective_runtime)

    mock_data.assert_called_once_with(configured_dir, default_dir,
                                      test_objective_runtime)
    mock_plot.assert_called_once_with(points, filename, **plot_params)
    assert figure_string == f"\\includegraphics[width=0.6\\textwidth]{{{filename}}}"


def test_get_timeouts(mocker: MockFixture) -> None:
    """Test get_timeouts correctly computes timeouts and overlapping values for dicts."""
    conf_dict = {
        "instance-1.cnf": 100.0,
        "instance-2.cnf": 100.0,
        "instance-3.cnf": 0.01,
        "instance-4.cnf": 0.01,
    }
    default_dict = {
        "instance-1.cnf": 0.01,
        "instance-2.cnf": 100.0,
        "instance-3.cnf": 100.0,
        "instance-4.cnf": 100.0,
    }
    cutoff = 10

    configured, default, overlap = sgrch.get_timeouts(
        conf_dict, default_dict, cutoff)

    assert configured == 2
    assert default == 3
    assert overlap == 1


def test_get_ablation_table(mocker: MockFixture) -> None:
    """Test get_ablation_table calls sah.get_ablation_table and transforms its string."""
    sah_ablation_table = (
        [["Round", "Flipped parameter", "Source value", "Target value",
          "Validation result"],
         ["0", "-source-", "N/A", "N/A", "76.53275"],
         ["1", "sel_var_div", "3", "6", "68.41392"],
         ["2", "-target-", "N/A", "N/A", "92.06944"]])
    mock_table = mocker.patch("sparkle.solver.ablation.AblationScenario."
                              "read_ablation_table",
                              return_value=sah_ablation_table)

    table_string = sgrch.get_ablation_table(ablation_scenario)

    mock_table.assert_called_once()
    assert table_string == (r"\begin{tabular}{rp{0.25\linewidth}rrr}"
                            r"\textbf{Round} & \textbf{Flipped parameter} & "
                            r"\textbf{Source value} & \textbf{Target value} & "
                            r"\textbf{Validation result} \\ \hline "
                            r"0 & -source- & N/A & N/A & 76.53275 \\ "
                            r"1 & sel_var_div & 3 & 6 & 68.41392 \\ "
                            r"2 & -target- & N/A & N/A & 92.06944 \\ "
                            r"\end{tabular}")


def test_get_dict_variable_to_value_with_test(mocker: MockFixture) -> None:
    """Test get_dict_variable_to_value returns correct dictionary.

    If a test instance is present, the function should add the corresponding entry
    to the dictionary.
    """
    train_instance = "train-instance"
    test_instance = "test-instance"
    output_dir = Path("configuration/report")
    common_dict = {
        "common-1": "1",
        "common-2": "2",
        "featuresBool": r"\featuresfalse"
    }
    test_dict = {
        "test-1": "3",
        "test-2": "4"
    }
    mock_common = mocker.patch("sparkle.platform.generate_report_for_configuration."
                               "get_dict_variable_to_value_common",
                               return_value=common_dict)
    mock_test = mocker.patch("sparkle.platform.generate_report_for_configuration."
                             "get_dict_variable_to_value_test",
                             return_value=test_dict)

    full_dict = sgrch.configuration_report_variables(
        Path("configuration/report"), solver, configurator, validator, Path(),
        Path(), train_instance, 1, test_instance, None)

    mock_common.assert_called_once_with(solver, configurator, validator,
                                        None, Path(), train_instance,
                                        output_dir)
    mock_test.assert_called_once_with(output_dir, solver, configurator, validator,
                                      None, train_instance, test_instance)
    assert full_dict == {
        "testBool": r"\testtrue",
        "ablationBool": r"\ablationfalse"
    } | common_dict | test_dict


def test_configuration_report_variables_without_test(mocker: MockFixture) -> None:
    """Test get_dict_variable_to_value returns correct dictionary.

    If no test instance is present, the function should add the corresponding entry
    to the dictionary.
    """
    train_instance = "train-instance"
    test_instance = None
    output_dir = Path("configuration/report")
    common_dict = {
        "common-1": "1",
        "common-2": "2",
        "featuresBool": r"\featuresfalse"
    }
    mock_common = mocker.patch("sparkle.platform.generate_report_for_configuration."
                               "get_dict_variable_to_value_common",
                               return_value=common_dict)

    full_dict = sgrch.configuration_report_variables(
        output_dir, solver, configurator, validator, Path(),
        Path(), train_instance, 1, test_instance, None)

    mock_common.assert_called_once_with(solver, configurator, validator, None, Path(),
                                        train_instance, output_dir)
    assert full_dict == {
        "testBool": r"\testfalse",
        "ablationBool": r"\ablationfalse"
    } | common_dict


def test_configuration_report_variables_with_ablation(mocker: MockFixture) -> None:
    """Test get_dict_variable_to_value returns correct dictionary.

    If `ablation` is set to True, the key `ablationBool` should not be set in the
    dictionary.
    """
    train_instance = "train-instance"
    test_instance = "test-instance"
    output_dir = Path("configuration/report")
    common_dict = {
        "common-1": "1",
        "common-2": "2",
        "featuresBool": r"\featuresfalse"
    }
    test_dict = {
        "test-1": "3",
        "test-2": "4"
    }
    mock_common = mocker.patch("sparkle.platform.generate_report_for_configuration."
                               "get_dict_variable_to_value_common",
                               return_value=common_dict)
    mock_test = mocker.patch("sparkle.platform.generate_report_for_configuration."
                             "get_dict_variable_to_value_test",
                             return_value=test_dict)

    full_dict = sgrch.configuration_report_variables(
        output_dir, solver, configurator, validator, Path(),
        Path(), train_instance, 1, test_instance, ablation_scenario)

    mock_common.assert_called_once_with(
        solver, configurator, validator, ablation_scenario,
        Path(), train_instance, output_dir)
    mock_test.assert_called_once_with(
        output_dir, solver, configurator, validator,
        ablation_scenario, train_instance, test_instance)
    assert full_dict == {
        "testBool": r"\testtrue"
    } | common_dict | test_dict


def test_configuration_report_variables_with_features(mocker: MockFixture) -> None:
    """Test get_dict_variable_to_value returns correct dictionary.

    If the key `featuresBool` in the common dictionary is found and set to true,
    the corresponding other keys should be added to the final dictionary.
    """
    train_instance = "train-instance"
    test_instance = "test-instance"
    output_dir = Path("configuration/report")
    common_dict = {
        "common-1": "1",
        "common-2": "2",
        "featuresBool": r"\featurestrue"
    }
    test_dict = {
        "test-1": "3",
        "test-2": "4"
    }
    mock_common = mocker.patch("sparkle.platform.generate_report_for_configuration."
                               "get_dict_variable_to_value_common",
                               return_value=common_dict)
    mock_test = mocker.patch("sparkle.platform.generate_report_for_configuration."
                             "get_dict_variable_to_value_test",
                             return_value=test_dict)
    mocker.patch("pathlib.Path.iterdir",
                 return_value=[Path("extract1"), Path("extract2")])
    mocker.patch("sparkle.platform.latex.list_to_latex", return_value="43")
    mocker.patch("pathlib.Path.mkdir", return_value=None)
    extractor_dir = Path("extract/dir")
    full_dict = sgrch.configuration_report_variables(
        output_dir, solver, configurator, validator, extractor_dir,
        Path(), train_instance, 1, test_instance, ablation_scenario)

    mock_common.assert_called_once_with(
        solver, configurator, validator, ablation_scenario, Path(), train_instance,
        output_dir)
    mock_test.assert_called_once_with(
        output_dir, solver, configurator, validator, ablation_scenario, train_instance,
        test_instance)
    assert full_dict == {
        "testBool": r"\testtrue",
        "numFeatureExtractors": "2",
        "featureExtractorList": "43",
        "featureComputationCutoffTime": "44"
    } | common_dict | test_dict


def test_get_dict_variable_to_value_common(mocker: MockFixture) -> None:
    """Test get_dict_variable_to_value_common creates the correct dictionary.

    Test that all needed functions are called to retrieve values and that these
    values are added to the common dictionary.
    """
    train_instance = Path("train-instance")
    validation_data = [
        ["SolverName", "{}", "InstanceSetName", "InstanceName", "STATUS", "0", "25.323"]]
    report_dir = "reports/directory"
    cutoff = 60
    configurator.scenario.cutoff_time = cutoff
    configurator.scenario.number_of_runs = 25
    configurator.scenario.wallclock_time = 600
    configurator.scenario.sparkle_objective = test_objective_quality

    mocker.patch("sparkle.about.version", "0.8")
    mocker.patch("sparkle.platform.generate_"
                 "report_for_configuration."
                 "get_average_performance",
                 side_effect=[42.1, 42.2])
    mocker.patch("sparkle.platform.generate_report_for_configuration."
                 "get_features_bool",
                 return_value="\\featuresfalse")
    mock_figure = mocker.patch("sparkle.platform.generate_"
                               "report_for_configuration."
                               "get_figure_configured_vs_default_on_"
                               "instance_set",
                               return_value="figure-string")
    mock_timeouts = mocker.patch("sparkle.platform.generate_"
                                 "report_for_configuration."
                                 "get_timeouts_instanceset",
                                 return_value=(2, 3, 1))
    mock_ablation_bool = mocker.patch("sparkle.platform.generate_"
                                      "report_for_configuration."
                                      "get_ablation_bool",
                                      return_value="ablationtrue")
    mock_ablation_table = mocker.patch("sparkle.platform.generate_"
                                       "report_for_configuration."
                                       "get_ablation_table",
                                       return_value="ablation/path")
    mocker.patch("sparkle.configurator.implementations.smac2."
                 "SMAC2.get_optimal_configuration", return_value=(0, "123"))
    mocker.patch("pathlib.Path.iterdir", return_value=[Path("test1")])
    mocker.patch("sparkle.solver.validator.Validator.get_validation_results",
                 return_value=validation_data)
    bib_path = Path("tex/bib.bib")
    common_dict = sgrch.get_dict_variable_to_value_common(
        solver, configurator, validator, ablation_scenario, bib_path, train_instance,
        report_dir)

    mock_figure.assert_called_once_with(solver, train_instance.name, validation_data,
                                        validation_data, report_dir, "QUALITY",
                                        float(cutoff), test_objective_quality)
    mock_timeouts.assert_called_once_with(
        solver, train_instance, configurator, validator, 60)
    mock_ablation_bool.assert_called_once_with(ablation_scenario)
    mock_ablation_table.assert_called_once_with(ablation_scenario)

    assert common_dict == {
        "performanceMeasure": "ACCURACY",
        "runtimeBool": "\\runtimefalse",
        "solver": solver.name,
        "instanceSetTrain": train_instance.name,
        "sparkleVersion": "0.8",
        "numInstanceInTrainingInstanceSet": 1,
        "numSmacRuns": 25,
        "smacObjective": "QUALITY",
        "smacWholeTimeBudget": 600,
        "smacEachRunCutoffTime": cutoff,
        "optimisedConfiguration": "123",
        "optimisedConfigurationTrainingPerformancePAR": 42.1,
        "defaultConfigurationTrainingPerformancePAR": 42.2,
        "figure-configured-vs-default-train": "figure-string",
        "timeoutsTrainDefault": 3,
        "timeoutsTrainConfigured": 2,
        "timeoutsTrainOverlap": 1,
        "ablationBool": "ablationtrue",
        "ablationPath": "ablation/path",
        "bibliographypath": bib_path.absolute(),
        "featuresBool": "\\featuresfalse"
    }


def test_get_dict_variable_to_value_test(mocker: MockFixture) -> None:
    """Test get_dict_variable_to_value_test creates the correct dictionary.

    Test that all needed functions are called to retrieve values and that these
    values are added to the common dictionary.
    """
    train_set = instance_set(Path("tests/test_files/Instances/Train-Instance-Set"))
    test_set = instance_set(Path("tests/test_files/Instances/Test-Instance-Set"))
    validation_data = [
        ["SolverName", "{}", "InstanceSetName", "InstanceName", "STATUS", "0", "25.323"]]
    cutoff = "60"

    mocker.patch("sparkle.platform.generate_"
                 "report_for_configuration."
                 "get_average_performance",
                 side_effect=[42.1, 42.2])
    mock_figure = mocker.patch("sparkle.platform.generate_"
                               "report_for_configuration."
                               "get_figure_configured_vs_default_on_"
                               "instance_set",
                               return_value="figure-string")
    mock_timeouts = mocker.patch("sparkle.platform.generate_"
                                 "report_for_configuration."
                                 "get_timeouts_instanceset",
                                 return_value=(2, 3, 1))
    mock_ablation_bool = mocker.patch("sparkle.platform.generate_"
                                      "report_for_configuration."
                                      "get_ablation_bool",
                                      return_value="ablationtrue")
    mock_ablation_table = mocker.patch("sparkle.platform.generate_"
                                       "report_for_configuration."
                                       "get_ablation_table",
                                       return_value="ablation/path")
    mocker.patch("pathlib.Path.iterdir", return_value=[Path("test1")])
    mocker.patch("sparkle.solver.validator.Validator.get_validation_results",
                 return_value=validation_data)
    mocker.patch("sparkle.configurator.implementations.SMAC2."
                 "get_optimal_configuration", return_value=(0.0, "configurino"))
    configurator.scenario.cutoff_time = 60
    configurator.scenario.sparkle_objective = test_objective_quality
    test_dict = sgrch.get_dict_variable_to_value_test(Path("configuration/report"),
                                                      solver,
                                                      configurator,
                                                      validator,
                                                      ablation_scenario,
                                                      train_set,
                                                      test_set)

    mock_figure.assert_called_once_with(
        solver, test_set.name, validation_data, validation_data,
        Path("configuration/report"), "QUALITY", float(cutoff),
        test_objective_quality, data_type="test")
    mock_timeouts.assert_called_once_with(
        solver, test_set, configurator, validator, 60)
    mock_ablation_bool.assert_called_once_with(ablation_scenario)
    mock_ablation_table.assert_called_once_with(ablation_scenario)
    assert test_dict == {
        "instanceSetTest": test_set.name,
        "numInstanceInTestingInstanceSet": 1,
        "optimisedConfigurationTestingPerformancePAR": 42.1,
        "defaultConfigurationTestingPerformancePAR": 42.2,
        "figure-configured-vs-default-test": "figure-string",
        "timeoutsTestDefault": 3,
        "timeoutsTestConfigured": 2,
        "timeoutsTestOverlap": 1,
        "ablationBool": "ablationtrue",
        "ablationPath": "ablation/path"
    }


def test_generate_report_for_configuration_train(mocker: MockFixture) -> None:
    """Test generate_report_for_configuration_train generates report.

    The function should call functions to prepare report generation and call
    `generate_report_for_configuration_common` with the right parameters.
    """
    train_instance = "train-instance"

    value_dict = {
        "key-1": "value-1",
        "key-2": "value-2",
        "bibliographypath": ""
    }

    report_dir = Path("configuration/report")

    mock_dict = mocker.patch("sparkle.platform.generate_report_for_configuration."
                             "configuration_report_variables",
                             return_value=value_dict)
    mock_generate_report = mocker.patch("sparkle.platform.latex.generate_report",
                                        return_value=None)

    sgrch.generate_report_for_configuration(solver,
                                            configurator,
                                            validator,
                                            Path(),
                                            report_dir,
                                            Path(),
                                            Path(),
                                            1.0,
                                            train_instance,
                                            ablation=True)
    mock_dict.assert_called_once_with(report_dir, solver,
                                      configurator, validator, Path(),
                                      Path(), 1.0, train_instance, None, True)
    mock_generate_report.assert_called_once()


def test_generate_report_for_configuration(mocker: MockFixture) -> None:
    """Test generate_report_for_configuration generates report.

    The function should call functions to prepare report generation and call
    `generate_report_for_configuration_common` with the right parameters.
    """
    train_instance = "train-instance"
    test_instance = "test-instance"
    ablation = True

    value_dict = {
        "key-1": "value-1",
        "key-2": "value-2"
    }

    report_dir = Path("configuration/report")

    mock_dict = mocker.patch("sparkle.platform.generate_report_for_configuration."
                             "configuration_report_variables",
                             return_value=value_dict)
    mock_generate_report = mocker.patch("sparkle.platform.latex.generate_report",
                                        return_value=None)

    sgrch.generate_report_for_configuration(solver,
                                            configurator,
                                            validator,
                                            Path(),
                                            report_dir,
                                            Path(),
                                            Path(),
                                            1.0,
                                            train_instance,
                                            test_instance, ablation)

    mock_dict.assert_called_once_with(
        report_dir, solver, configurator, validator,
        Path(), Path(), 1.0, train_instance, test_instance, ablation)
    mock_generate_report.assert_called_once()



"""Classes and Enums to control settings."""
from __future__ import annotations
import configparser
from enum import Enum
from pathlib import Path
from pathlib import PurePath

from sparkle.types import SparkleObjective, resolve_objective
from sparkle.types.objective import PAR
from sparkle.solver import Selector
from sparkle.configurator.configurator import Configurator
from sparkle.solver.verifier import SATVerifier
from sparkle.configurator import implementations as cim

from runrunner import Runner
from sparkle.platform.cli_types import VerbosityLevel


class SettingState(Enum):
    """Enum of possible setting states."""

    NOT_SET = 0
    DEFAULT = 1
    FILE = 2
    CMD_LINE = 3


class Settings:
    """Class to read, write, set, and get settings."""
    # CWD Prefix
    cwd_prefix = Path()  # Empty for now

    # Library prefix
    lib_prefix = Path(__file__).parent.parent.resolve()

    # Default directory names
    rawdata_dir = Path("Raw_Data")
    analysis_dir = Path("Analysis")
    __settings_dir = Path("Settings")
    __settings_file = Path("sparkle_settings.ini")

    # Default settings path
    DEFAULT_settings_path = PurePath(cwd_prefix / __settings_dir / __settings_file)

    # Default library pathing
    DEFAULT_components = lib_prefix / "Components"

    # Example settings path
    DEFAULT_example_settings_path = PurePath(DEFAULT_components / "sparkle_settings.ini")

    # Runsolver component
    DEFAULT_runsolver_dir = DEFAULT_components / "runsolver" / "src"
    DEFAULT_runsolver_exec = DEFAULT_runsolver_dir / "runsolver"

    # Ablation component
    DEFAULT_ablation_dir = DEFAULT_components / "ablationAnalysis-0.9.4"
    DEFAULT_ablation_exec = DEFAULT_ablation_dir / "ablationAnalysis"
    DEFAULT_ablation_validation_exec = DEFAULT_ablation_dir / "ablationValidation"

    # Autofolio component
    DEFAULT_general_sparkle_selector = DEFAULT_components / "AutoFolio/scripts/autofolio"

    # Report component
    DEFAULT_latex_source = DEFAULT_components / "Sparkle-latex-source"
    DEFAULT_latex_bib = DEFAULT_latex_source / "SparkleReport.bib"

    # Default input directory pathing
    DEFAULT_solver_dir = cwd_prefix / "Solvers"
    DEFAULT_instance_dir = cwd_prefix / "Instances"
    DEFAULT_extractor_dir = cwd_prefix / "Extractors"
    DEFAULT_snapshot_dir = cwd_prefix / "Snapshots"

    # Default output directory pathing
    DEFAULT_tmp_output = cwd_prefix / "Tmp"
    DEFAULT_output = cwd_prefix / "Output"
    DEFAULT_configuration_output = DEFAULT_output / "Configuration"
    DEFAULT_selection_output = DEFAULT_output / "Selection"
    DEFAULT_validation_output = DEFAULT_output / "Validation"
    DEFAULT_parallel_portfolio_output = DEFAULT_output / "Parallel_Portfolio"
    DEFAULT_ablation_output = DEFAULT_output / "Ablation"
    DEFAULT_log_output = DEFAULT_output / "Log"

    # Default output subdirs
    DEFAULT_configuration_output_raw = DEFAULT_configuration_output / rawdata_dir
    DEFAULT_configuration_output_analysis = DEFAULT_configuration_output / analysis_dir
    DEFAULT_selection_output_raw = DEFAULT_selection_output / rawdata_dir
    DEFAULT_selection_output_analysis = DEFAULT_selection_output / analysis_dir
    DEFAULT_parallel_portfolio_output_raw =\
        DEFAULT_parallel_portfolio_output / rawdata_dir
    DEFAULT_parallel_portfolio_output_analysis =\
        DEFAULT_parallel_portfolio_output / analysis_dir

    # Old default output dirs which should be part of something else
    DEFAULT_feature_data = DEFAULT_output / "Feature_Data"
    DEFAULT_performance_data = DEFAULT_output / "Performance_Data"

    # Collection of all working dirs for platform
    DEFAULT_working_dirs = [
        DEFAULT_output, DEFAULT_configuration_output,
        DEFAULT_selection_output, DEFAULT_validation_output,
        DEFAULT_tmp_output,
        DEFAULT_log_output,
        DEFAULT_solver_dir, DEFAULT_instance_dir,
        DEFAULT_feature_data, DEFAULT_performance_data,
        DEFAULT_extractor_dir,
    ]

    # Old default file paths from GV which should be turned into variables
    DEFAULT_feature_data_path =\
        DEFAULT_feature_data / "feature_data.csv"
    DEFAULT_performance_data_path =\
        DEFAULT_performance_data / "performance_data.csv"

    # Constant default values
    DEFAULT_general_sparkle_objective = PAR(10)
    DEFAULT_general_sparkle_configurator = cim.SMAC2.__name__
    DEFAULT_general_solution_verifier = str(None)
    DEFAULT_general_target_cutoff_time = 60
    DEFAULT_general_extractor_cutoff_time = 60
    DEFAULT_number_of_jobs_in_parallel = 25
    DEFAULT_general_verbosity = VerbosityLevel.STANDARD
    DEFAULT_general_check_interval = 10

    DEFAULT_config_wallclock_time = 600
    DEFAULT_config_cpu_time = None
    DEFAULT_config_solver_calls = None
    DEFAULT_config_number_of_runs = 25
    DEFAULT_configurator_target_cutoff_length = "max"

    DEFAULT_portfolio_construction_timeout = None

    DEFAULT_slurm_max_parallel_runs_per_node = 8

    DEFAULT_ablation_racing = False

    DEFAULT_parallel_portfolio_check_interval = 4
    DEFAULT_parallel_portfolio_num_seeds_per_solver = 1

    def __init__(self: Settings, file_path: PurePath = None) -> None:
        """Initialise a settings object."""
        # Settings 'dictionary' in configparser format
        self.__settings = configparser.ConfigParser()

        # Setting flags
        self.__general_sparkle_objective_set = SettingState.NOT_SET
        self.__general_sparkle_configurator_set = SettingState.NOT_SET
        self.__general_sparkle_selector_set = SettingState.NOT_SET
        self.__general_solution_verifier_set = SettingState.NOT_SET
        self.__general_target_cutoff_time_set = SettingState.NOT_SET
        self.__general_extractor_cutoff_time_set = SettingState.NOT_SET
        self.__general_verbosity_set = SettingState.NOT_SET
        self.__general_check_interval_set = SettingState.NOT_SET

        self.__config_wallclock_time_set = SettingState.NOT_SET
        self.__config_cpu_time_set = SettingState.NOT_SET
        self.__config_solver_calls_set = SettingState.NOT_SET
        self.__config_number_of_runs_set = SettingState.NOT_SET

        self.__run_on_set = SettingState.NOT_SET
        self.__number_of_jobs_in_parallel_set = SettingState.NOT_SET
        self.__slurm_max_parallel_runs_per_node_set = SettingState.NOT_SET
        self.__configurator_target_cutoff_length_set = SettingState.NOT_SET
        self.__slurm_extra_options_set = dict()
        self.__ablation_racing_flag_set = SettingState.NOT_SET

        self.__parallel_portfolio_check_interval_set = SettingState.NOT_SET
        self.__parallel_portfolio_num_seeds_per_solver_set = SettingState.NOT_SET

        self.__general_sparkle_configurator = None

        if file_path is None:
            # Initialise settings from default file path
            self.read_settings_ini()
        else:
            # Initialise settings from a given file path
            self.read_settings_ini(file_path)

    def read_settings_ini(self: Settings, file_path: PurePath = DEFAULT_settings_path,
                          state: SettingState = SettingState.FILE) -> None:
        """Read the settings from an INI file."""
        # Read file
        file_settings = configparser.ConfigParser()
        file_settings.read(file_path)

        # Set internal settings based on data read from FILE if they were read
        # successfully
        if file_settings.sections() != []:
            section = "general"
            option_names = ("objective", )
            for option in option_names:
                if file_settings.has_option(section, option):
                    value = [resolve_objective(obj) for obj in
                             file_settings.get(section, option).split(",")]
                    self.set_general_sparkle_objectives(value, state)
                    file_settings.remove_option(section, option)

            # Comma so python understands it's a tuple...
            option_names = ("configurator", )
            for option in option_names:
                if file_settings.has_option(section, option):
                    value = file_settings.get(section, option)
                    self.set_general_sparkle_configurator(value, state)
                    file_settings.remove_option(section, option)

            option_names = ("selector", )
            for option in option_names:
                if file_settings.has_option(section, option):
                    value = file_settings.get(section, option)
                    self.set_general_sparkle_selector(value, state)
                    file_settings.remove_option(section, option)

            option_names = ("solution_verifier", )
            for option in option_names:
                if file_settings.has_option(section, option):
                    value = file_settings.get(section, option).lower()
                    self.set_general_solution_verifier(value, state)
                    file_settings.remove_option(section, option)

            option_names = ("target_cutoff_time",
                            "cutoff_time_each_solver_call")
            for option in option_names:
                if file_settings.has_option(section, option):
                    value = file_settings.getint(section, option)
                    self.set_general_target_cutoff_time(value, state)
                    file_settings.remove_option(section, option)

            option_names = ("extractor_cutoff_time",
                            "cutoff_time_each_feature_computation")
            for option in option_names:
                if file_settings.has_option(section, option):
                    value = file_settings.getint(section, option)
                    self.set_general_extractor_cutoff_time(value, state)
                    file_settings.remove_option(section, option)

            option_names = ("run_on", )
            for option in option_names:
                if file_settings.has_option(section, option):
                    value = file_settings.get(section, option)
                    self.set_run_on(value, state)
                    file_settings.remove_option(section, option)

            option_names = ("verbosity", )
            for option in option_names:
                if file_settings.has_option(section, option):
                    value = VerbosityLevel.from_string(
                        file_settings.get(section, option))
                    self.set_general_verbosity(value, state)
                    file_settings.remove_option(section, option)

            option_names = ("check_interval", )
            for option in option_names:
                if file_settings.has_option(section, option):
                    value = int(file_settings.get(section, option))
                    self.set_general_check_interval(value, state)
                    file_settings.remove_option(section, option)

            section = "configuration"
            option_names = ("wallclock_time", )
            for option in option_names:
                if file_settings.has_option(section, option):
                    value = file_settings.getint(section, option)
                    self.set_config_wallclock_time(value, state)
                    file_settings.remove_option(section, option)

            option_names = ("cpu_time", )
            for option in option_names:
                if file_settings.has_option(section, option):
                    value = file_settings.getint(section, option)
                    self.set_config_cpu_time(value, state)
                    file_settings.remove_option(section, option)

            option_names = ("solver_calls", )
            for option in option_names:
                if file_settings.has_option(section, option):
                    value = file_settings.getint(section, option)
                    self.set_config_solver_calls(value, state)
                    file_settings.remove_option(section, option)

            option_names = ("number_of_runs", )
            for option in option_names:
                if file_settings.has_option(section, option):
                    value = file_settings.getint(section, option)
                    self.set_config_number_of_runs(value, state)
                    file_settings.remove_option(section, option)

            option_names = ("target_cutoff_length", "smac_each_run_cutoff_length")
            for option in option_names:
                if file_settings.has_option(section, option):
                    value = file_settings.get(section, option)
                    self.set_configurator_target_cutoff_length(value, state)
                    file_settings.remove_option(section, option)

            section = "slurm"
            option_names = ("number_of_jobs_in_parallel", "num_job_in_parallel")
            for option in option_names:
                if file_settings.has_option(section, option):
                    value = file_settings.getint(section, option)
                    self.set_number_of_jobs_in_parallel(value, state)
                    file_settings.remove_option(section, option)

            option_names = ("max_parallel_runs_per_node", "clis_per_node")
            for option in option_names:
                if file_settings.has_option(section, option):
                    value = file_settings.getint(section, option)
                    self.set_slurm_max_parallel_runs_per_node(value, state)
                    file_settings.remove_option(section, option)

            section = "ablation"
            option_names = ("racing", "ablation_racing")
            for option in option_names:
                if file_settings.has_option(section, option):
                    value = file_settings.getboolean(section, option)
                    self.set_ablation_racing_flag(value, state)
                    file_settings.remove_option(section, option)

            section = "parallel_portfolio"
            option_names = ("check_interval", )
            for option in option_names:
                if file_settings.has_option(section, option):
                    value = int(file_settings.get(section, option))
                    self.set_parallel_portfolio_check_interval(value, state)
                    file_settings.remove_option(section, option)

            option_names = ("num_seeds_per_solver", )
            for option in option_names:
                if file_settings.has_option(section, option):
                    value = int(file_settings.get(section, option))
                    self.set_parallel_portfolio_number_of_seeds_per_solver(value, state)
                    file_settings.remove_option(section, option)

            # TODO: Report on any unknown settings that were read
            sections = file_settings.sections()

            for section in sections:
                for option in file_settings[section]:
                    # TODO: Should check the options are valid Slurm options
                    if section == "slurm":
                        value = file_settings.get(section, option)
                        self.add_slurm_extra_option(option, value, state)
                    else:
                        print(f'Unrecognised section - option combination: "{section} '
                              f'{option}" in file {file_path} ignored')

        # Print error if unable to read the settings
        else:
            print(f"ERROR: Failed to read settings from {file_path} The file may have "
                  "been empty, located in a different path, or be in another format than"
                  " INI. Default Settings values be used.")

    def write_used_settings(self: Settings) -> None:
        """Write the used settings to the default locations."""
        # Write to latest settings file
        self.write_settings_ini(self.__settings_dir / "latest.ini")

    def write_settings_ini(self: Settings, file_path: Path) -> None:
        """Write the settings to an INI file."""
        # Create needed directories if they don't exist
        file_path.parent.mkdir(parents=True, exist_ok=True)
        slurm_extra_section_options = None
        if self.__settings.has_section("slurm_extra"):
            # Slurm extra options are not written as a seperate section
            slurm_extra_section_options = {}
            for key in self.__settings["slurm_extra"]:
                self.__settings["slurm"][key] = self.__settings["slurm_extra"][key]
                slurm_extra_section_options[key] = self.__settings["slurm_extra"][key]
            self.__settings.remove_section("slurm_extra")
        # Write the settings to file
        with file_path.open("w") as settings_file:
            self.__settings.write(settings_file)
        # Rebuild slurm extra if needed
        if slurm_extra_section_options is not None:
            self.__settings.add_section("slurm_extra")
            for key in slurm_extra_section_options:
                self.__settings["slurm_extra"][key] = slurm_extra_section_options[key]

    def __init_section(self: Settings, section: str) -> None:
        if section not in self.__settings:
            self.__settings[section] = {}

    @staticmethod
    def __check_setting_state(current_state: SettingState,
                              new_state: SettingState, name: str) -> bool:
        change_setting_ok = True

        if current_state == SettingState.FILE and new_state == SettingState.DEFAULT:
            change_setting_ok = False
            print(f"Warning: Attempting to overwrite setting for {name} with default "
                  "value; keeping the value read from file!")
        elif (current_state == SettingState.CMD_LINE
              and new_state == SettingState.DEFAULT):
            change_setting_ok = False
            print(f"Warning: Attempting to overwrite setting for {name} with default "
                  "value; keeping the value read from command line!")
        elif current_state == SettingState.CMD_LINE and new_state == SettingState.FILE:
            change_setting_ok = False
            print(f"Warning: Attempting to overwrite setting for {name} with value from "
                  "file; keeping the value read from command line!")

        return change_setting_ok

    # General settings ###
    def set_general_sparkle_objectives(
            self: Settings,
            value: list[SparkleObjective] = [DEFAULT_general_sparkle_objective, ],
            origin: SettingState = SettingState.DEFAULT) -> None:
        """Set the sparkle objective."""
        section = "general"
        name = "objective"
        if value is not None and self.__check_setting_state(
                self.__general_sparkle_objective_set, origin, name):
            if isinstance(value, list):
                value = ",".join([str(obj) for obj in value])
            else:
                value = str(value)
            # Append standard Sparkle Objectives
            if "status" not in value:
                value += ",status"
            if "cpu_time" not in value:
                value += ",cpu_time"
            if "wall_time" not in value:
                value += ",wall_time"
            if "memory" not in value:
                value += ",memory"
            self.__init_section(section)
            self.__general_sparkle_objective_set = origin
            self.__settings[section][name] = value

    def get_general_sparkle_objectives(self: Settings) -> list[SparkleObjective]:
        """Return the performance measure."""
        if self.__general_sparkle_objective_set == SettingState.NOT_SET:
            self.set_general_sparkle_objectives()

        return [resolve_objective(obj)
                for obj in self.__settings["general"]["objective"].split(",")]

    def set_general_sparkle_configurator(
            self: Settings,
            value: str = DEFAULT_general_sparkle_configurator,
            origin: SettingState = SettingState.DEFAULT) -> None:
        """Set the Sparkle configurator."""
        section = "general"
        name = "configurator"
        if value is not None and self.__check_setting_state(
                self.__general_sparkle_configurator_set, origin, name):
            self.__init_section(section)
            self.__general_sparkle_configurator_set = origin
            self.__settings[section][name] = value

    def get_general_sparkle_configurator(self: Settings) -> Configurator:
        """Return the configurator init method."""
        if self.__general_sparkle_configurator_set == SettingState.NOT_SET:
            self.set_general_sparkle_configurator()
        if self.__general_sparkle_configurator is None:
            configurator_subclass =\
                cim.resolve_configurator(self.__settings["general"]["configurator"])
            if configurator_subclass is not None:
                self.__general_sparkle_configurator = configurator_subclass(
                    objectives=self.get_general_sparkle_objectives(),
                    base_dir=Settings.DEFAULT_tmp_output,
                    output_path=Settings.DEFAULT_configuration_output_raw)
            else:
                print("WARNING: Configurator class name not recognised:"
                      f'{self.__settings["general"]["configurator"]}. '
                      "Configurator not set.")
        return self.__general_sparkle_configurator

    def set_general_sparkle_selector(
            self: Settings,
            value: Path = DEFAULT_general_sparkle_selector,
            origin: SettingState = SettingState.DEFAULT) -> None:
        """Set the Sparkle selector."""
        section = "general"
        name = "selector"
        if value is not None and self.__check_setting_state(
                self.__general_sparkle_selector_set, origin, name):
            self.__init_section(section)
            self.__general_sparkle_selector_set = origin
            self.__settings[section][name] = str(value)

    def get_general_sparkle_selector(self: Settings) -> Selector:
        """Return the selector init method."""
        if self.__general_sparkle_selector_set == SettingState.NOT_SET:
            self.set_general_sparkle_selector()
        return Selector(Path(self.__settings["general"]["selector"]),
                        self.DEFAULT_selection_output_raw)

    def set_general_solution_verifier(
            self: Settings, value: str = DEFAULT_general_solution_verifier,
            origin: SettingState = SettingState.DEFAULT) -> None:
        """Set the solution verifier to use."""
        section = "general"
        name = "solution_verifier"

        if value is not None and self.__check_setting_state(
                self.__general_solution_verifier_set, origin, name):
            self.__init_section(section)
            self.__general_solution_verifier_set = origin
            self.__settings[section][name] = value

    def get_general_solution_verifier(self: Settings) -> object:
        """Return the solution verifier to use."""
        if self.__general_solution_verifier_set == SettingState.NOT_SET:
            self.set_general_solution_verifier()
        name = self.__settings["general"]["solution_verifier"].lower()
        if name == str(SATVerifier()).lower():
            return SATVerifier()
        return None

    def set_general_target_cutoff_time(
            self: Settings, value: int = DEFAULT_general_target_cutoff_time,
            origin: SettingState = SettingState.DEFAULT) -> None:
        """Set the cutoff time in seconds for target algorithms."""
        section = "general"
        name = "target_cutoff_time"

        if value is not None and self.__check_setting_state(
                self.__general_target_cutoff_time_set, origin, name):
            self.__init_section(section)
            self.__general_target_cutoff_time_set = origin
            self.__settings[section][name] = str(value)

    def get_general_target_cutoff_time(self: Settings) -> int:
        """Return the cutoff time in seconds for target algorithms."""
        if self.__general_target_cutoff_time_set == SettingState.NOT_SET:
            self.set_general_target_cutoff_time()
        return int(self.__settings["general"]["target_cutoff_time"])

    def set_general_extractor_cutoff_time(
            self: Settings, value: int = DEFAULT_general_extractor_cutoff_time,
            origin: SettingState = SettingState.DEFAULT) -> None:
        """Set the cutoff time in seconds for feature extraction."""
        section = "general"
        name = "extractor_cutoff_time"

        if value is not None and self.__check_setting_state(
                self.__general_extractor_cutoff_time_set, origin, name):
            self.__init_section(section)
            self.__general_extractor_cutoff_time_set = origin
            self.__settings[section][name] = str(value)

    def get_general_extractor_cutoff_time(self: Settings) -> int:
        """Return the cutoff time in seconds for feature extraction."""
        if self.__general_extractor_cutoff_time_set == SettingState.NOT_SET:
            self.set_general_extractor_cutoff_time()
        return int(self.__settings["general"]["extractor_cutoff_time"])

    def set_number_of_jobs_in_parallel(
            self: Settings, value: int = DEFAULT_number_of_jobs_in_parallel,
            origin: SettingState = SettingState.DEFAULT) -> None:
        """Set the number of runs Sparkle can do in parallel."""
        section = "slurm"
        name = "number_of_jobs_in_parallel"

        if value is not None and self.__check_setting_state(
                self.__number_of_jobs_in_parallel_set, origin, name):
            self.__init_section(section)
            self.__number_of_jobs_in_parallel_set = origin
            self.__settings[section][name] = str(value)

    def get_number_of_jobs_in_parallel(self: Settings) -> int:
        """Return the number of runs Sparkle can do in parallel."""
        if self.__number_of_jobs_in_parallel_set == SettingState.NOT_SET:
            self.set_number_of_jobs_in_parallel()

        return int(self.__settings["slurm"]["number_of_jobs_in_parallel"])

    def set_general_verbosity(
            self: Settings, value: VerbosityLevel = DEFAULT_general_verbosity,
            origin: SettingState = SettingState.DEFAULT) -> None:
        """Set the general verbosity to use."""
        section = "general"
        name = "verbosity"

        if value is not None and self.__check_setting_state(
                self.__general_verbosity_set, origin, name):
            self.__init_section(section)
            self.__general_verbosity_set = origin
            self.__settings[section][name] = value.name

    def get_general_verbosity(self: Settings) -> VerbosityLevel:
        """Return the general verbosity."""
        if self.__general_verbosity_set == SettingState.NOT_SET:
            self.set_general_verbosity()

        return VerbosityLevel.from_string(
            self.__settings["general"]["verbosity"])

    def set_general_check_interval(
            self: Settings,
            value: int = DEFAULT_general_check_interval,
            origin: SettingState = SettingState.DEFAULT) -> None:
        """Set the general check interval."""
        section = "general"
        name = "check_interval"

        if value is not None and self.__check_setting_state(
                self.__general_check_interval_set, origin, name):
            self.__init_section(section)
            self.__general_check_interval_set = origin
            self.__settings[section][name] = str(value)

    def get_general_check_interval(self: Settings) -> int:
        """Return the general check interval."""
        if self.__general_check_interval_set == SettingState.NOT_SET:
            self.set_general_check_interval()

        return int(
            self.__settings["general"]["check_interval"])

    # Configuration settings ###

    def set_config_wallclock_time(
            self: Settings, value: int = DEFAULT_config_wallclock_time,
            origin: SettingState = SettingState.DEFAULT) -> None:
        """Set the budget per configuration run in seconds (wallclock)."""
        section = "configuration"
        name = "wallclock_time"

        if value is not None and self.__check_setting_state(
                self.__config_wallclock_time_set, origin, name):
            self.__init_section(section)
            self.__config_wallclock_time_set = origin
            self.__settings[section][name] = str(value)

    def get_config_wallclock_time(self: Settings) -> int:
        """Return the budget per configuration run in seconds (wallclock)."""
        if self.__config_wallclock_time_set == SettingState.NOT_SET:
            self.set_config_wallclock_time()
        return int(self.__settings["configuration"]["wallclock_time"])

    def set_config_cpu_time(
            self: Settings, value: int = DEFAULT_config_cpu_time,
            origin: SettingState = SettingState.DEFAULT) -> None:
        """Set the budget per configuration run in seconds (cpu)."""
        section = "configuration"
        name = "cpu_time"

        if value is not None and self.__check_setting_state(
                self.__config_cpu_time_set, origin, name):
            self.__init_section(section)
            self.__config_cpu_time_set = origin
            self.__settings[section][name] = str(value)

    def get_config_cpu_time(self: Settings) -> int | None:
        """Return the budget per configuration run in seconds (cpu)."""
        if self.__config_cpu_time_set == SettingState.NOT_SET:
            self.set_config_cpu_time()
            return None

        return int(self.__settings["configuration"]["cpu_time"])

    def set_config_solver_calls(
            self: Settings, value: int = DEFAULT_config_solver_calls,
            origin: SettingState = SettingState.DEFAULT) -> None:
        """Set the number of solver calls."""
        section = "configuration"
        name = "solver_calls"

        if value is not None and self.__check_setting_state(
                self.__config_solver_calls_set, origin, name):
            self.__init_section(section)
            self.__config_solver_calls_set = origin
            self.__settings[section][name] = str(value)

    def get_config_solver_calls(self: Settings) -> int | None:
        """Return the number of solver calls."""
        if self.__config_solver_calls_set == SettingState.NOT_SET:
            self.set_config_solver_calls()
            return None

        return int(self.__settings["configuration"]["solver_calls"])

    def set_config_number_of_runs(
            self: Settings, value: int = DEFAULT_config_number_of_runs,
            origin: SettingState = SettingState.DEFAULT) -> None:
        """Set the number of configuration runs."""
        section = "configuration"
        name = "number_of_runs"

        if value is not None and self.__check_setting_state(
                self.__config_number_of_runs_set, origin, name):
            self.__init_section(section)
            self.__config_number_of_runs_set = origin
            self.__settings[section][name] = str(value)

    def get_config_number_of_runs(self: Settings) -> int:
        """Return the number of configuration runs."""
        if self.__config_number_of_runs_set == SettingState.NOT_SET:
            self.set_config_number_of_runs()

        return int(self.__settings["configuration"]["number_of_runs"])

    # Configuration: SMAC specific settings ###

    def set_configurator_target_cutoff_length(
            self: Settings, value: str = DEFAULT_configurator_target_cutoff_length,
            origin: SettingState = SettingState.DEFAULT) -> None:
        """Set the target algorithm cutoff length."""
        section = "configuration"
        name = "target_cutoff_length"

        if value is not None and self.__check_setting_state(
                self.__configurator_target_cutoff_length_set, origin, name):
            self.__init_section(section)
            self.__configurator_target_cutoff_length_set = origin
            self.__settings[section][name] = str(value)

    def get_configurator_target_cutoff_length(self: Settings) -> str:
        """Return the target algorithm cutoff length."""
        if self.__configurator_target_cutoff_length_set == SettingState.NOT_SET:
            self.set_configurator_target_cutoff_length()
        return self.__settings["configuration"]["target_cutoff_length"]

    # Slurm settings ###

    def set_slurm_max_parallel_runs_per_node(
            self: Settings,
            value: int = DEFAULT_slurm_max_parallel_runs_per_node,
            origin: SettingState = SettingState.DEFAULT) -> None:
        """Set the number of algorithms Slurm can run in parallel per node."""
        section = "slurm"
        name = "max_parallel_runs_per_node"

        if value is not None and self.__check_setting_state(
                self.__slurm_max_parallel_runs_per_node_set, origin, name):
            self.__init_section(section)
            self.__slurm_max_parallel_runs_per_node_set = origin
            self.__settings[section][name] = str(value)

    def get_slurm_max_parallel_runs_per_node(self: Settings) -> int:
        """Return the number of algorithms Slurm can run in parallel per node."""
        if self.__slurm_max_parallel_runs_per_node_set == SettingState.NOT_SET:
            self.set_slurm_max_parallel_runs_per_node()

        return int(self.__settings["slurm"]["max_parallel_runs_per_node"])

    # SLURM extra options

    def add_slurm_extra_option(self: Settings, name: str, value: str,
                               origin: SettingState = SettingState.DEFAULT) -> None:
        """Add additional Slurm options."""
        section = "slurm_extra"

        current_state = (self.__slurm_extra_options_set[name]
                         if name in self.__slurm_extra_options_set
                         else SettingState.NOT_SET)

        if value is not None and self.__check_setting_state(current_state, origin, name):
            self.__init_section(section)
            self.__slurm_extra_options_set[name] = origin
            self.__settings[section][name] = str(value)

    def get_slurm_extra_options(self: Settings,
                                as_args: bool = False) -> dict | list:
        """Return a dict with additional Slurm options."""
        section = "slurm_extra"
        options = dict()

        if "slurm_extra" in self.__settings.sections():
            for option in self.__settings["slurm_extra"]:
                options[option] = self.__settings.get(section, option)
        if as_args:
            return [f"--{key}={options[key]}" for key in options.keys()]
        return options

    # Ablation settings ###

    def set_ablation_racing_flag(self: Settings, value: bool = DEFAULT_ablation_racing,
                                 origin: SettingState = SettingState.DEFAULT) -> None:
        """Set a flag indicating whether racing should be used for ablation."""
        section = "ablation"
        name = "racing"

        if value is not None and self.__check_setting_state(
                self.__ablation_racing_flag_set, origin, name):
            self.__init_section(section)
            self.__ablation_racing_flag_set = origin
            self.__settings[section][name] = str(value)

    def get_ablation_racing_flag(self: Settings) -> bool:
        """Return a bool indicating whether the racing flag is set for ablation."""
        if self.__ablation_racing_flag_set == SettingState.NOT_SET:
            self.set_ablation_racing_flag()

        return bool(self.__settings["ablation"]["racing"])

    # Parallel Portfolio settings

    def set_parallel_portfolio_check_interval(
            self: Settings,
            value: int = DEFAULT_parallel_portfolio_check_interval,
            origin: SettingState = SettingState.DEFAULT) -> None:
        """Set the parallel portfolio check interval."""
        section = "parallel_portfolio"
        name = "check_interval"

        if value is not None and self.__check_setting_state(
                self.__parallel_portfolio_check_interval_set, origin, name):
            self.__init_section(section)
            self.__parallel_portfolio_check_interval_set = origin
            self.__settings[section][name] = str(value)

    def get_parallel_portfolio_check_interval(self: Settings) -> int:
        """Return the parallel portfolio check interval."""
        if self.__parallel_portfolio_check_interval_set == SettingState.NOT_SET:
            self.set_parallel_portfolio_check_interval()

        return int(
            self.__settings["parallel_portfolio"]["check_interval"])

    def set_parallel_portfolio_number_of_seeds_per_solver(
            self: Settings,
            value: int = DEFAULT_parallel_portfolio_num_seeds_per_solver,
            origin: SettingState = SettingState.DEFAULT) -> None:
        """Set the parallel portfolio seeds per solver to start."""
        section = "parallel_portfolio"
        name = "num_seeds_per_solver"

        if value is not None and self.__check_setting_state(
                self.__parallel_portfolio_num_seeds_per_solver_set, origin, name):
            self.__init_section(section)
            self.__parallel_portfolio_num_seeds_per_solver_set = origin
            self.__settings[section][name] = str(value)

    def get_parallel_portfolio_number_of_seeds_per_solver(self: Settings) -> int:
        """Return the parallel portfolio seeds per solver to start."""
        if self.__parallel_portfolio_num_seeds_per_solver_set == SettingState.NOT_SET:
            self.set_parallel_portfolio_number_of_seeds_per_solver()

        return int(
            self.__settings["parallel_portfolio"]["num_seeds_per_solver"])

    def set_run_on(self: Settings, value: Runner = str,
                   origin: SettingState = SettingState.DEFAULT) -> None:
        """Set the compute on which to run."""
        section = "general"
        name = "run_on"

        if value is not None and self.__check_setting_state(
                self.__run_on_set, origin, name):
            self.__init_section(section)
            self.__run_on_set = origin
            self.__settings[section][name] = value

    def get_run_on(self: Settings) -> Runner:
        """Return the compute on which to run."""
        return Runner(self.__settings["general"]["run_on"])

    @staticmethod
    def check_settings_changes(cur_settings: Settings, prev_settings: Settings) -> bool:
        """Check if there are changes between the previous and the current settings.

        Prints any section changes, printing None if no setting was found.

        Args:
          cur_settings: The current settings
          prev_settings: The previous settings

        Returns:
          True iff there are no changes.
        """
        cur_dict = cur_settings.__settings._sections
        prev_dict = prev_settings.__settings._sections

        cur_sections_set = set(cur_dict.keys())
        prev_sections_set = set(prev_dict.keys())

        sections_removed = prev_sections_set - cur_sections_set
        if sections_removed:
            print("Warning: the following sections have been removed:")
            for section in sections_removed:
                print(f"  - Section '{section}'")

        sections_added = cur_sections_set - prev_sections_set
        if sections_added:
            print("Warning: the following sections have been added:")
            for section in sections_added:
                print(f"  - Section '{section}'")

        sections_remained = cur_sections_set & prev_sections_set
        option_changed = False
        for section in sections_remained:
            printed_section = False
            names = set(cur_dict[section].keys()) | set(prev_dict[section].keys())
            for name in names:
                # if name is not present in one of the two dicts, get None as placeholder
                cur_val = cur_dict[section].get(name, None)
                prev_val = prev_dict[section].get(name, None)
                if cur_val != prev_val:
                    # Have we printed the initial warning?
                    if not option_changed:
                        print("Warning: The following attributes/options have changed:")
                        option_changed = True

                    # do we have yet to print the section?
                    if not printed_section:
                        print(f"  - In the section '{section}':")
                        printed_section = True

                    # print actual change
                    print(f"    · '{name}' changed from '{prev_val}' to '{cur_val}'")

        return not (sections_removed or sections_added or option_changed)



#!/usr/bin/env python3
"""Sparkle output structures."""
from __future__ import annotations
import sys
from pathlib import Path

from runrunner.base import Status

from sparkle.solver import Solver
from sparkle.types import SolverStatus, SparkleObjective
from sparkle.instance import InstanceSet
from sparkle.structures import PerformanceDataFrame


class ValidationResults:
    """Class that stores validation information and results."""
    def __init__(self: ValidationResults, solver: Solver,
                 configuration: dict, instance_set: InstanceSet,
                 results: list[list[str, Status, float, float]]) -> None:
        """Initalize ValidationResults.

        Args:
            solver: The name of the solver
            configuration: The configuration being used
            instance_set: The set of instances
            results: Validation results in the format:
                [["instance", "status", "quality", "runtime"]]
        """
        self.solver = solver
        self.configuration = configuration
        self.instance_set = instance_set
        self.result_header = ["instance", "status", "quality", "runtime"]
        self.result_vals = results


class ConfigurationResults:
    """Class that aggregates configuration results."""
    def __init__(self: ConfigurationResults, metrics: float,
                 results: ValidationResults) -> None:
        """Initalize ConfigurationResults.

        Args:
            metrics: The performance of a configured solver
            results: The results for one configuration
        """
        self.performance = metrics
        self.results = results


class SelectionSolverData:
    """Class that stores solver information."""
    def __init__(self: SelectionSolverData,
                 solver_performance_ranking: list[tuple[str, float]],
                 num_solvers: int) -> None:
        """Initalize SelectionSolverData.

        Args:
            solver_performance_ranking: list with solvers ranked by avg. performance
            num_solvers: The number of solvers
        """
        self.solver_performance_ranking = solver_performance_ranking
        self.single_best_solver = solver_performance_ranking[0][0]
        self.num_solvers = num_solvers


class SelectionPerformance:
    """Class that stores selection performance results."""
    def __init__(self: SelectionSolverData,
                 performance_path: Path,
                 vbs_performance: float,
                 objective: SparkleObjective) -> None:
        """Initalize SelectionPerformance.

        Args:
            performance_path: Path to portfolio selector performance
            vbs_performance: The performance of the virtual best selector
            objective: The objective (Performance type)
        """
        if not performance_path.exists():
            print(f"ERROR: {performance_path} does not exist.")
            sys.exit(-1)
        actual_performance_data = PerformanceDataFrame(performance_path)
        self.vbs_performance = vbs_performance
        self.actual_performance = actual_performance_data.mean(
            objective=objective.name)
        self.metric = objective.name


class ParallelPortfolioResults:
    """Class that stores parallel portfolio results."""
    def __init__(self: ParallelPortfolioResults,
                 unsolved_instances: int,
                 sbs: str,
                 runtime_solvers: dict[str, float],
                 instance_results: dict[str, list]) -> None:
        """Initalize SelectionSolverData.

        Args:
            unsolved_instances: Number of unsolved instances
            sbs: Name of the single best solver
            runtime_solvers: Dictionary containing penalised average runtime per solver
            instance_results: Dictionary containing
        """
        self.unsolved_instances = unsolved_instances
        self.sbs = sbs
        self.runtime_solvers = runtime_solvers
        self.instance_results = instance_results

        self.solver_performance = {}
        # Iterate over each instance and aggregate the results
        for _, results in self.instance_results.items():
            for solver_result in results:
                solver_name = solver_result[0]
                outcome = solver_result[1]
                # Initialize the solver's record in solver_performance if not present
                if solver_name not in self.solver_performance:
                    self.solver_performance[solver_name] = {
                        status: 0 for status in SolverStatus
                    }
                # Increment the appropriate outcome count
                self.solver_performance[solver_name][outcome] += 1



import numpy as np
import logging

from ConfigSpace.configuration_space import Configuration
from aslib_scenario.aslib_scenario import ASlibScenario

from autofolio.autofolio import AutoFolio

__author__ = "Marius Lindauer"
__license__ = "BSD"
__version__ = "2.1.0"


class AFCsvFacade(object):

    def __init__(self,
                 perf_fn:str,
                 feat_fn:str,
                 objective:str = "solution_quality",
                 runtime_cutoff:float = None,
                 maximize:bool = True,
                 cv_fn:str = None,
                 seed: int = 12345
                 ):
        """ Constructor """
        self.scenario = ASlibScenario()
        self.scenario.read_from_csv(perf_fn=perf_fn,
                                               feat_fn=feat_fn,
                                               objective=objective,
                                               runtime_cutoff=runtime_cutoff,
                                               maximize=maximize,
                                               cv_fn=cv_fn)
        self.seed = seed

        self.af = AutoFolio(random_seed=seed)
        self.logger = logging.getLogger("AF Facade")

    def fit(self,
            config:Configuration=None,
            save_fn:str = None):
        """ Train AutoFolio on data from init"""
        self.logger.info("Fit")
        if config is None:
            cs = self.af.get_cs(self.scenario, {})
            config = cs.get_default_configuration()
        feature_pre_pipeline, pre_solver, selector = self.af.fit(scenario=self.scenario, config=config)

        if save_fn:
            self.af._save_model(save_fn, self.scenario, feature_pre_pipeline, pre_solver, selector, config)
        self.logger.info("AutoFolio model saved to %s" %(save_fn))

    def tune(self,
             wallclock_limit:int = 1200,
             runcount_limit:int = np.inf,
             ):

        config = self.af.get_tuned_config(self.scenario,
                                       wallclock_limit=wallclock_limit,
                                       runcount_limit=runcount_limit,
                                       autofolio_config={},
                                       seed=self.seed)
        self.logger.info("Optimized Configuration: %s" %(config))
        return config

    def cross_validation(self, config:Configuration):
        """ run a cross validation on given AutoFolio configuration"""
        score = -1 * self.af.run_cv(config=config, scenario=self.scenario, folds=int(self.scenario.cv_data.max().max()))
        self.logger.info("AF's final performance %f" %(score))

        return score

    @staticmethod
    def load_and_predict(vec: np.ndarray,
                         load_fn:str):
        """ get predicted algorithm for given meta-feature vector"""
        af = AutoFolio(random_seed=42) # random seed doesn't matter here
        pred = af.read_model_and_predict(model_fn=load_fn, feature_vec=vec)
        print("Selected Schedule [(algorithm, budget)]: %s" % (pred))
        return pred[0][0]






"""Types definition for features to unify extractors across the Sparkle platform."""
from __future__ import annotations
from enum import Enum


class FeatureGroup(str, Enum):
    """Various feature groups."""
    BASE = "base"
    PRE = "pre"
    DIAMETER = "dia"
    LOBJOIS = "lobjois"
    UNIT = "unit"
    SP = "sp"
    LS = "ls"
    LP = "lp"
    CL = "cl"


class FeatureSubgroup(str, Enum):
    """Various feature subgroups. Only used for embedding in with feature names."""
    VCG = "Variable-Clause-Graph"
    POSNEG = "Postive-Negative-Literals"
    HORNY = "Horn-Formula"
    VG = "Variable-Graph"
    KLB = "Kevin-Leyton-Brown"
    CG = "Clause-Graph"
    BIAS = "bias"
    UNCONSTRAINT = "unconstraint"
    NUM = "num"
    SIZE = "size"


class FeatureType(str, Enum):
    """Various feature types."""
    MIN = "min"
    MAX = "max"
    MEAN = "mean"
    ENTROPY = "entropy"
    QUANTILE_10 = "q10"
    QUANTILE_25 = "q25"
    QUANTILE_50 = "q50"
    QUANTILE_75 = "q75"
    QUANTILE_90 = "q90"
    COEFFICIENT_VARIATION = "coefficient_of_variation"
    NUMBER_OF_VARS_ORIGINAL = "n_vars_original"
    NUMBER_OF_CLAUSES_ORIGINAL = "n_clauses_original"
    NUMBER_OF_VARS = "n_vars"
    NUMBER_OF_CLAUSES = "n_clauses"
    REDUCED_VARS = "reduced_vars"
    REDUCED_CLAUSES = "reduced_clauses"
    PRE_FEATURE_TIME = "pre_featuretime"
    VARS_CLAUSES_RATIO = "vars_clauses_ratio"
    CLAUSE_RATIO_MEAN = "clause_ratio_mean"
    CLAUSE_RATIO_MINIMUM = "clause_ratio_minimum"
    CLAUSE_RATIO_MAXIMUM = "clause_ratio_maximum"
    CLAUSE_RATIO_ENTROPY = "clause_ratio_entropy"
    CLAUSE_MEAN = "clause_mean"
    CLAUSE_MIN = "clause_min"
    CLAUSE_MAX = "clause_max"
    CLAUSE_FRACTION = "clauses_fraction"
    CLAUSE_ENTROPY = "clause_entropy"
    CLAUSE_COEFFICIENT_VARIATION = "clause_coefficient_of_variation"
    VAR_MEAN = "variable_mean"
    VAR_MIN = "variable_min"
    VAR_MAX = "variable_max"
    VAR_STD = "variable_standard_deviation"
    VAR_ENTROPY = "variable_entropy"
    VAR_COEFFICIENT_VARIATION = "variable_coefficient_of_variation"
    CLUSTER_COEFFICIENT_MEAN = "cluster_coefficient_mean"
    CLUSTER_COEFFICIENT_VARIATION = "cluster_coefficient_of_variation"
    CLUSTER_COEFFICIENT_MIN = "cluster_coefficient_min"
    CLUSTER_COEFFICIENT_MAX = "cluster_coefficient_max"
    CLUSTER_COEFFICIENT_ENTROPY = "cluster_coefficient_entropy"
    UNARY = "unary"
    BINARY = "binary"
    TRINARY = "trinary"
    FEATURE_TIME = "feature_time"

    @staticmethod
    def with_subgroup(subgroup: FeatureSubgroup, feature: FeatureType) -> str:
        """Return a standardised string with a subgroup embedded."""
        return f"{subgroup.value}_{feature.value}"



"""Test public methods of sparkle performance data csv."""

from __future__ import annotations
from unittest import TestCase
from pathlib import Path

import pytest

from sparkle.structures import PerformanceDataFrame


class TestPerformanceData(TestCase):
    """Class for testing performance data object."""

    def setUp(self: TestPerformanceData) -> None:
        """Create csv objects from files for the tests."""
        self.csv_example_path =\
            Path("tests/test_files/performance/example-runtime-performance.csv")
        self.pd = PerformanceDataFrame(self.csv_example_path)
        self.csv_example_with_nan_path =\
            Path("tests/test_files/performance/"
                 "example-runtime-performance-with-empty.csv")
        self.pd_nan = PerformanceDataFrame(self.csv_example_with_nan_path)
        self.csv_example_mo = Path("tests/test_files/performance/example_data_MO.csv")
        self.pd_mo = PerformanceDataFrame(self.csv_example_mo)

    def test_get_job_list(self: TestPerformanceData) -> None:
        """Test job list method, without and with recompute bool."""
        job_list = []
        result = self.pd.get_job_list()
        assert result == job_list

        job_list = [("Instance1", 1, "AlgorithmA"), ("Instance1", 1, "AlgorithmB"),
                    ("Instance1", 1, "AlgorithmC"), ("Instance1", 1, "AlgorithmD"),
                    ("Instance1", 1, "AlgorithmE"), ("Instance2", 1, "AlgorithmA"),
                    ("Instance2", 1, "AlgorithmB"), ("Instance2", 1, "AlgorithmC"),
                    ("Instance2", 1, "AlgorithmD"), ("Instance2", 1, "AlgorithmE"),
                    ("Instance3", 1, "AlgorithmA"), ("Instance3", 1, "AlgorithmB"),
                    ("Instance3", 1, "AlgorithmC"), ("Instance3", 1, "AlgorithmD"),
                    ("Instance3", 1, "AlgorithmE"), ("Instance4", 1, "AlgorithmA"),
                    ("Instance4", 1, "AlgorithmB"), ("Instance4", 1, "AlgorithmC"),
                    ("Instance4", 1, "AlgorithmD"), ("Instance4", 1, "AlgorithmE"),
                    ("Instance5", 1, "AlgorithmA"), ("Instance5", 1, "AlgorithmB"),
                    ("Instance5", 1, "AlgorithmC"), ("Instance5", 1, "AlgorithmD"),
                    ("Instance5", 1, "AlgorithmE")]
        result = self.pd.get_job_list(rerun=True)
        assert result == job_list

    def test_num_instances(self: TestPerformanceData) -> None:
        """Test the number of instances getter method."""
        num_instances = 5
        assert self.pd.num_instances == num_instances
        assert self.pd_nan.num_instances == num_instances
        num_instances = 2
        assert self.pd_mo.num_instances == num_instances

    def test_get_list_remaining_performance_computation_job(self: TestPerformanceData)\
            -> None:
        """Test get remaining performance computation job getter."""
        remaining = {}
        result = self.pd.remaining_jobs()
        assert result == remaining

        remaining = {"Instance1": {"AlgorithmA"}, "Instance2": {"AlgorithmA"},
                     "Instance3": {"AlgorithmC", "AlgorithmA"},
                     "Instance4": {"AlgorithmA", "AlgorithmE"},
                     "Instance5": {"AlgorithmA"}}
        result = self.pd_nan.remaining_jobs()
        assert result == remaining

        remaining = {}
        result = self.pd_mo.remaining_jobs()
        assert result == remaining

    def test_calc_best_performance_instance(self: TestPerformanceData)\
            -> None:
        """Test calculating best score on instance."""
        bp_instance_runtime = [30.0, 5.0, 3.0, 8.0, 41.0]
        result_min = self.pd.best_instance_performance()
        for idx in range(self.pd.num_instances):
            assert result_min.iloc[idx] == bp_instance_runtime[idx]

        bp_instance_accuracy = [0.930, 0.819]
        result_acc = self.pd_mo.best_instance_performance(objective="TrainAccuracy:max")
        for idx in range(self.pd_mo.num_instances):
            assert result_acc.iloc[idx] == bp_instance_accuracy[idx]

        bp_instance_val_accuracy = [0.88, 0.596]
        result_acc = self.pd_mo.best_instance_performance(
            objective="ValidationAccuracy:max")
        for idx in range(self.pd_mo.num_instances):
            assert result_acc.iloc[idx] == bp_instance_val_accuracy[idx]

    def test_calc_best_performance(self: TestPerformanceData) -> None:
        """Test calculating vbs on the entire portfolio."""
        vbs_portfolio = 17.4
        result = self.pd.best_performance()
        assert result == vbs_portfolio

        vbs_portfolio = 0.738
        results = self.pd_mo.best_performance(objective="ValidationAccuracy:max")
        assert results == vbs_portfolio

        vbs_portfolio = 4.449999999999999
        results = self.pd_mo.best_performance(objective="PAR10")
        assert results == vbs_portfolio

    def test_marginal_contribution(self: TestPerformanceData) -> None:
        """Test marginal contribution."""
        marginal = [("AlgorithmA", 0.0, 17.4),
                    ("AlgorithmB", 1.4252873563218393, 24.8),
                    ("AlgorithmC", 1.1264367816091956, 19.6),
                    ("AlgorithmD", 0.0, 17.4), ("AlgorithmE", 1.7471264367816093, 30.4)]
        result = self.pd.marginal_contribution()
        assert result == marginal

        marginal = [("RandomForest", 0.0, 0.738),
                    ("MultiLayerPerceptron", 0.8699186991869919, 0.642)]
        result = self.pd_mo.marginal_contribution(objective="ValidationAccuracy:max")
        assert result == marginal

        marginal = [("RandomForest", 8.786516853932584, 39.099999999999994),
                    ("MultiLayerPerceptron", 0.0, 4.449999999999999)]
        result = self.pd_mo.marginal_contribution(objective="PAR10")
        assert result == marginal

    @pytest.mark.filterwarnings("ignore::FutureWarning")
    def test_get_solver_ranking(self: TestPerformanceData) -> None:
        """Test getting the solver ranking list with penalty."""
        rank_list = [("AlgorithmB", 41.0), ("AlgorithmC", 43.6),
                     ("AlgorithmE", 52.6), ("AlgorithmD", 54.8), ("AlgorithmA", 55.0)]
        result = self.pd.get_solver_ranking()
        assert result == rank_list

        rank_list = [("MultiLayerPerceptron", 0.691734),
                     ("RandomForest", 0.5930799999999999)]
        result = self.pd_mo.get_solver_ranking(objective="ValidationAccuracy:max")
        assert result == rank_list

        rank_list = [("RandomForest", 4.9079999999999995),
                     ("MultiLayerPerceptron", 102.24799999999999)]
        result = self.pd_mo.get_solver_ranking(objective="PAR10")
        assert result == rank_list



#!/usr/bin/env python2.7
# encoding: utf-8

'''
emptyWrapper -- template for a wrapper based in genericWrapper.py

@author:     Marius Lindauer, Chris Fawcett, Alex Fréchette, Frank Hutter
@copyright:  2014 AClib. All rights reserved.
@license:    GPL
@contact:    lindauer@informatik.uni-freiburg.de, fawcettc@cs.ubc.ca, afrechet@cs.ubc.ca, fh@informatik.uni-freiburg.de

'''

from genericWrapper import AbstractWrapper

class EmptyWrapper(AbstractWrapper):
    
    def get_command_line_args(self, runargs, config):
        '''
        Returns the command line call string to execute the target algorithm (here: Spear).
        Args:
            runargs: a map of several optional arguments for the execution of the target algorithm.
                    {
                      "instance": <instance>,
                      "specifics" : <extra data associated with the instance>,
                      "cutoff" : <runtime cutoff>,
                      "runlength" : <runlength cutoff>,
                      "seed" : <seed>
                    }
            config: a mapping from parameter name to parameter value
        Returns:
            A command call list to execute the target algorithm.
        '''
        binary_path = "<TODO>"
        cmd = "%s %s" %(binary_path, runargs["instance"])       
        return cmd
    
    def process_results(self, filepointer, out_args):
        '''
        Parse a results file to extract the run's status (SUCCESS/CRASHED/etc) and other optional results.
    
        Args:
            filepointer: a pointer to the file containing the solver execution standard out.
            out_args : a map with {"exit_code" : exit code of target algorithm} 
        Returns:
            A map containing the standard AClib run results. The current standard result map as of AClib 2.06 is:
            {
                "status" : <"SUCCESS"/"SAT"/"UNSAT"/"TIMEOUT"/"CRASHED"/"ABORT">,
                "runtime" : <runtime of target algrithm>,
                "quality" : <a domain specific measure of the quality of the solution [optional]>,
                "misc" : <a (comma-less) string that will be associated with the run [optional]>
            }
            ATTENTION: The return values will overwrite the measured results of the runsolver (if runsolver was used). 
        '''
        resultMap = {}
        return resultMap

if __name__ == "__main__":
    wrapper = EmptyWrapper()
    wrapper.main()



"""This package provides platform support for Sparkle."""
from sparkle.platform.cli_types import CommandName, COMMAND_DEPENDENCIES
from sparkle.platform.settings_objects import Settings, SettingState



#!/usr/bin/env python3
"""Sparkle command to wait for one or more other commands to complete execution."""
import sys
import signal
import time
import argparse
from pathlib import Path

from runrunner.slurm import SlurmRun
from runrunner.base import Status
from tabulate import tabulate

from sparkle.platform.cli_types import VerbosityLevel, TEXT
from sparkle.CLI.help import logging
from sparkle.CLI.help import argparse_custom as ac
from sparkle.CLI.help import global_variables as gv


def parser_function() -> argparse.ArgumentParser:
    """Define the command line arguments.

    Returns:
      The argument parser.
    """
    parser = argparse.ArgumentParser()
    parser.add_argument(*ac.JobIDsArgument.names, **ac.JobIDsArgument.kwargs)
    return parser


def get_runs_from_file(path: Path, print_error: bool = False) -> list[SlurmRun]:
    """Retrieve all run objects from file storage.

    Args:
        path: Path object where to look recursively for the files.

    Returns:
        List of all found SlumRun objects.
    """
    runs = []
    for file in path.rglob("*.json"):
        # TODO: RunRunner should be adapted to have more general methods for runs
        # So this method can work for both local and slurm
        try:
            run_obj = SlurmRun.from_file(file)
            runs.append(run_obj)
        except Exception as ex:
            # Not a (correct) RunRunner JSON file
            if print_error:
                print(f"[WARNING] Could not load file: {file}. Exception: {ex}")
    return runs


def wait_for_jobs(path: Path,
                  check_interval: int,
                  verbosity: VerbosityLevel = VerbosityLevel.STANDARD,
                  filter: list[str] = None) -> None:
    """Wait for all active jobs to finish executing.

    Args:
        path: The Path where to look for the stored jobs.
        check_interval: The time in seconds between updating the jobs.
        verbosity: Amount of information shown.
            The lower verbosity means lower computational load.
        filter: If present, only show the given job ids.
    """
    # Filter jobs on relevant status
    jobs = [run for run in get_runs_from_file(path)
            if run.status == Status.WAITING or run.status == Status.RUNNING]

    if filter is not None:
        jobs = [job for job in jobs if job.run_id in filter]

    running_jobs = jobs

    def signal_handler(num: int, _: any) -> None:
        """Create clean exit for CTRL + C."""
        if num == signal.SIGINT:
            sys.exit(0)

    signal.signal(signal.SIGINT, signal_handler)
    # If verbosity is quiet there is no need for further information
    if verbosity == VerbosityLevel.QUIET:
        prev_jobs = len(running_jobs) + 1
        while len(running_jobs) > 0:
            if len(running_jobs) < prev_jobs:
                print(f"Waiting for {len(running_jobs)} jobs...", flush=True)
            time.sleep(check_interval)
            prev_jobs = len(running_jobs)
            running_jobs = [run for run in running_jobs
                            if run.status == Status.WAITING
                            or run.status == Status.RUNNING]

    # If verbosity is standard the command will print a table with relevant information
    elif verbosity == VerbosityLevel.STANDARD:
        # Order in which to display the jobs
        status_order = {Status.COMPLETED: 0, Status.RUNNING: 1, Status.WAITING: 2}
        while len(running_jobs) > 0:
            # Information to be printed to the table
            information = [["RunId", "Name", "Partition", "Status",
                            "Dependencies", "Finished Jobs", "Run Time"]]
            running_jobs = [run for run in running_jobs
                            if run.status == Status.WAITING
                            or run.status == Status.RUNNING]
            sorted_jobs = sorted(
                jobs, key=lambda job: (status_order.get(job.status, 4), job.run_id))
            for job in sorted_jobs:
                # Count number of jobs that have finished
                finished_jobs_count = sum(1 for status in job.all_status
                                          if status == Status.COMPLETED)
                # Format job.status
                status_text = \
                    TEXT.format_text([TEXT.BOLD], job.status) \
                    if job.status == Status.RUNNING else \
                    (TEXT.format_text([TEXT.ITALIC], job.status)
                        if job.status == Status.COMPLETED else job.status)
                information.append(
                    [job.run_id,
                     job.name,
                     job.partition,
                     status_text,
                     "None" if len(job.dependencies) == 0
                        else ", ".join(job.dependencies),
                     f"{finished_jobs_count}/{len(job.all_status)}",
                     job.runtime])
            # Print the table
            table = tabulate(information, headers="firstrow", tablefmt="grid")
            print(table)
            time.sleep(check_interval)

            # Clears the table for the new table to be printed
            lines = table.count("\n") + 1
            # \033 is the escape character (ESC),
            # [{lines}A is the escape sequence that moves the cursor up.
            print(f"\033[{lines}A", end="")
            # [J is the escape sequence that clears the console from the cursor down
            print("\033[J", end="")

    print("All jobs done!")


if __name__ == "__main__":
    # Log command call
    logging.log_command(sys.argv)

    # Define command line arguments
    parser = parser_function()

    # Process command line arguments
    args = parser.parse_args()

    check_interval = gv.settings().get_general_check_interval()
    verbosity = gv.settings().get_general_verbosity()

    wait_for_jobs(path=gv.settings().DEFAULT_log_output,
                  check_interval=check_interval,
                  verbosity=verbosity,
                  filter=args.job_ids)



"""The Parameter Configuration Space Parser class."""
from __future__ import annotations
import re
import sys
import numpy as np
from enum import Enum
from abc import ABC
from pathlib import Path


class PCSObject(ABC):
    """General data structure to keep the pcs file in.

    Fields are added by functions, such that checks can be conducted.
    """
    def __init__(self: PCSObject) -> None:
        """Initialize the PCSObject."""
        self.params = []

    def add_param(self: PCSObject,
                  name: str,
                  structure: str = "integer",
                  domain: list = [-sys.maxsize, sys.maxsize],
                  scale: str = "linear",
                  default: str = "0",
                  comment: str = None) -> None:
        """Add a parameter to the PCSObject."""
        if structure not in ["integer", "real", "categorical", "ordinal"]:
            raise ValueError(f"Parameter structure {structure} not supported.")

        # Domain check
        if structure in ["integer", "real"]:
            if len(domain) != 2:
                raise ValueError(f"Parameter domain {domain} not supported.")
            pass
        elif structure == "categorical":
            # TODO: check categories
            scale = None

        self.params.append({
            "name": name,
            "structure": structure,
            "domain": domain,
            "scale": scale,
            "default": default,
            "comment": comment,
            "type": "parameter",
        })

    def add_constraint(self: PCSObject, **kwargs: any) -> None:
        """Add a constraint to the PCSObject."""
        # TODO add checks
        self.params.append({**kwargs, "type": "constraint"})

    def add_forbidden(self: PCSObject, **kwargs: any) -> None:
        """Add a forbidden clause to the PCSObject."""
        # TODO add checks
        self.params.append({**kwargs, "type": "forbidden"})

    def add_comment(self: PCSObject, **kwargs: any) -> None:
        """Add a comment to the PCSObject."""
        # TODO add checks
        self.params.append({**kwargs, "type": "comment"})

    def clear(self: PCSObject) -> None:
        """Clear the PCSObject."""
        self.params = []

    def get(self: PCSObject, name: str) -> dict:
        """Get a parameter from the PCSObject based on the name."""
        names = {p["name"]: i for i, p in enumerate(self.params) if "name" in p}
        if name in names:
            return self.params[names[name]]
        return None


class PCSConvention(Enum):
    """Internal pcs convention enum."""
    unknown = ""
    SMAC = "smac"
    ParamILS = "paramils"


class PCSParser(ABC):
    """Base interface object for the parser.

    It loads the pcs files into the generic pcs object. Once a parameter file is loaded,
    it can be exported to another file
    """

    def __init__(self: PCSParser, inherit: PCSParser = None) -> None:
        """Initialize the PCSParser."""
        if inherit is None:
            self.pcs = PCSObject()
        else:
            self.pcs = inherit.pcs

    @staticmethod
    def _format_string_to_enum(string: str) -> PCSConvention:
        """Convert string to PCSConvention."""
        for form in PCSConvention:
            if form.value == string:
                return form
        raise Exception("ERROR: parameter configuration space format is not supported.")

    def check_validity(self: PCSParser) -> bool:
        """Check the validity of the pcs."""
        # TODO implement

        # check if for all parameters in constraints and forbidden clauses exists
        # Check for conflict between default values and constraints and forbidden clauses
        return True

    def load(self: PCSParser, filepath: Path, convention: str = "smac") -> None:
        """Main import function."""
        if isinstance(filepath, str):
            filepath = Path(filepath)
        convention = self._format_string_to_enum(convention)

        # TODO check if file actually exists
        lines = filepath.open().readlines()
        if convention == PCSConvention.SMAC:
            parser = SMACParser(self)
            parser.parse(lines)
            self.pcs = parser.pcs
        else:
            raise Exception(f"ERROR: Importing the pcs convention for {convention.value}"
                            " is not yet implemented.")

    def export(self: PCSParser,
               convention: str = "smac",
               destination: Path = None) -> None:
        """Main export function."""
        convention = self._format_string_to_enum(convention)
        if convention == PCSConvention.ParamILS:
            pcs = ParamILSParser(self).compile()
        else:
            raise Exception(f"ERROR: Exporting the pcs convention for {convention.value}"
                            " is not yet implemented.")
        destination.open("w").write(pcs)


class SMACParser(PCSParser):
    """The SMAC parser class."""

    def parse(self: SMACParser, lines: list[str]) -> None:
        """Parse the pcs file."""
        self.pcs.clear()

        # PARAMS
        for line in lines:
            # The only forbidden characters in parameter names are:
            # spaces, commas, quotes, and parentheses
            regex = (r"(?P<name>[^\s\"',]*)\s+(?P<structure>\w*)\s+(?P<domain>(\[|\{)"
                     r".*(\]|\}))\s*\[(?P<default>.*)\]\s*(?P<scale>log)"
                     r"*\s*#*(?P<comment>.*)")
            m = re.match(regex, line)
            if m is not None:
                fields = m.groupdict()
                fields["domain"] = re.sub(r"(?:\[|\]|\{|\})", "", fields["domain"])
                fields["domain"] = re.split(r"\s*,\s*", fields["domain"])
                self.pcs.add_param(**fields)
                continue

            # CONSTRAINTS
            regex = (r"(?P<parameter>[^\s\"',]+)\s*\|\s"
                     r"*(?P<conditions>.+)\s*#*(?P<comment>.*)")
            m = re.match(regex, line)
            if m is not None:
                constraint = m.groupdict()
                constraint["conditions"] = self._parse_conditions(
                    constraint["conditions"])
                self.pcs.add_constraint(**constraint)
                continue

            # FORBIDDEN CLAUSES
            regex = r"\s*\{(?P<clauses>[^\}]+)\}\s*#*(?P<comment>.*)"
            m = re.match(regex, line)
            if m is not None:
                forbidden = m.groupdict()
                conditions = []
                # Simple clauses
                # {<parameter name 1>=<value 1>, ..., <parameter name N>=<value N>}
                if "," in forbidden["clauses"]:
                    forbidden["clause_type"] = "simple"
                    for clause in re.split(r"\s*,\s*", forbidden["clauses"]):
                        m = re.match(r"(?P<param>[^\s\"',=]+)\s*=\s*"
                                     r"(?P<value>[^\s\"',]+)", clause)
                        if m is not None:
                            conditions.append(m.groupdict())
                        else:
                            print(clause, "ERROR")

                else:  # Advanced clauses
                    forbidden["clause_type"] = "advanced"
                    # TODO decide if we need to further parse this down
                    conditions = [expr for expr in re.split(r"\s*(?:\|\||&&)\s*",
                                                            forbidden["clauses"])]

                if len(conditions) == 0:
                    raise Exception(f"ERROR: cannot parse the following line:\n'{line}'")

                forbidden["clauses"] = conditions

                self.pcs.add_forbidden(**forbidden)
                continue

            # COMMENTLINE
            regex = r"\s*#(?P<comment>.*)"
            m = re.match(regex, line)
            if m is not None:
                comment = m.groupdict()
                self.pcs.add_comment(**comment)
                continue

            # EMTPY LINE
            regex = r"^\s*$"
            m = re.match(regex, line)
            if m is not None:
                continue

            # RAISE ERROR
            raise Exception(f"ERROR: cannot parse the following line: \n'{line}'")

        return

    def _parse_conditions(self: SMACParser, conditions: str) -> list[tuple]:
        """Parse the conditions."""
        conditionlist = []
        condition = None
        operator = None
        nested = 0
        nested_start = 0
        condition_start = 0
        for pos, char in enumerate(conditions):
            # Nested clauses
            if char == "(":
                if nested == 0:
                    nested_start = pos
                nested += 1
            elif char == ")":
                nested -= 1
                if nested == 0:
                    condition = self._parse_conditions(conditions[nested_start + 1:pos])
                    conditionlist.append((operator, condition))
                    if (pos + 1) == len(conditions):
                        return conditionlist

            if pos > 1 and nested == 0:
                for op in ["||", "&&"]:
                    if conditions[pos - 1: pos + 1] == op:
                        if not isinstance(condition, list):
                            condition = self._parse_condition(
                                conditions[condition_start:pos - 1])
                            conditionlist.append((operator, condition))

                        operator = op
                        condition_start = pos + 1

        condition = self._parse_condition(conditions[condition_start:len(conditions)])
        conditionlist.append((operator, condition))

        return conditionlist

    @staticmethod
    def _parse_condition(condition: str) -> dict:
        """Parse the condition."""
        cont = False

        m = re.match(r"\s*(?P<parameter>[^\s\"',]+)\s*(?P<quantifier>==|!=|<|>|<=|>=)"
                     r"\s*(?P<value>[^\s\"',]+)\s*", condition)
        if m is not None:
            condition = {
                **m.groupdict(),
                "type": "numerical",
            }
            cont = True

        if not cont:
            m = re.match(r"\s*(?P<parameter>[^\s\"',]+)\s+"
                         r"in\s*\{(?P<items>[^\}]+)\}\s*", condition)
            if m is not None:
                condition = {
                    **m.groupdict(),
                    "type": "categorical",
                }
                condition["items"] = re.split(r",\s*", condition["items"])
            cont = True

        if not cont:
            raise Exception(f"ERROR: Couldn't parse '{condition}'")

        return condition

    def compile(self: SMACParser) -> str:
        """Compile the PCS."""
        # TODO implement
        pass


class ParamILSParser(PCSParser):
    """PCS parser for ParamILS format."""

    def parse(self: ParamILSParser, lines: list[str]) -> None:
        """Parse the PCS."""
        # TODO implement
        pass

    def compile(self: ParamILSParser) -> str:
        """Compile the PCS."""
        # TODO Produce warning if certain specifications cannot be kept in this format
        # TODO granularity parameter that sets how log and real ranges should be expanded
        granularity = 20

        lines = []
        for item in self.pcs.params:
            if item["type"] == "parameter":
                if item["structure"] in ["ordinal", "categorical"]:
                    domain = ",".join(item["domain"])
                elif item["structure"] == "integer":
                    if len(item["domain"]) != 2:
                        raise ValueError(f"Domain {item['domain']} not supported.")

                    (minval, maxval) = [int(i) for i in item["domain"]]
                    if item["scale"] != "log":
                        # domain = f"{minval}, {(minval + 1)}..{maxval}"
                        domain = list(np.linspace(minval, maxval, granularity))
                        domain = list(set(np.round(domain).astype(int)))  # Cast to int
                        if int(item["default"]) not in domain:
                            domain += [int(item["default"])]
                        domain.sort()

                        domain = ",".join([str(i) for i in domain])
                    else:
                        domain = list(np.unique(np.geomspace(minval, maxval, granularity,
                                                             dtype=int)))
                        # add default value
                        if int(item["default"]) not in domain:
                            domain += [int(item["default"])]
                            domain.sort()

                        domain = ",".join([str(i) for i in domain])

                elif item["structure"] == "real":
                    if len(item["domain"]) != 2:
                        raise ValueError(f"Domain {item['domain']} not supported.")

                    (minval, maxval) = [float(i) for i in item["domain"]]
                    if item["scale"] != "log":
                        domain = list(np.linspace(minval, maxval, granularity))
                    else:
                        domain = list(np.unique(np.geomspace(minval, maxval, granularity,
                                                             dtype=float)))
                    # add default value
                    if float(item["default"]) not in domain:
                        domain += [float(item["default"])]
                        domain.sort()

                    # Filter duplicated in string format
                    domain = list(set([f"{i}" for i in domain]))
                    domain.sort(key=float)
                    domain = ",".join(domain)

                domain = "{" + domain + "}"
                line = f"{item['name']} {domain} [{item['default']}]"
                if item["comment"] != "":
                    line += f" #{item['comment']}"

                lines.append(line)

        for item in self.pcs.params:
            if item["type"] == "constraint":
                line = f"{item['parameter']} | "
                line += self._compile_conditions(item["conditions"])
                if item["comment"] != "":
                    line += f" #{item['comment']}"
                lines.append(line)

        for item in self.pcs.params:
            if item["type"] == "forbidden":
                if item["clause_type"] == "simple":
                    clauses = [f"{cls['param']}={cls['value']}"
                               for cls in item["clauses"]]
                    line = "{" + ",".join(clauses) + "}"
                    if item["comment"] != "":
                        line += f"#{item['comment']}"
                    lines.append(line)
                else:
                    print("WARNING: Advanced forbidden clauses "
                          "are not supported by ParamILS.")
                pass

        lines = "\n".join(lines)
        return lines

    def _compile_conditions(self: ParamILSParser, conditions: list[tuple]) -> str:
        """Compile a list of conditions."""
        line = ""
        for operator, condition in conditions:
            if operator is not None:
                line += f" {operator} "

            if isinstance(condition, list):
                line += f"({self._compile_conditions(condition)})"
            else:
                if condition["type"] == "numerical":
                    line += f"{condition['parameter']} in " + "{"
                    param = self.pcs.get(condition["parameter"])
                    if param["structure"] == "categorical":
                        if condition["value"] in param["domain"]:
                            line += f"{condition['value']}" + "}"
                    # line += "{parameter} {quantifier} {value}".format(**condition)
                if condition["type"] == "categorical":
                    items = ", ".join(condition["items"])
                    line += f"{condition['parameter']} in {{{items}}}"
        return line
